# ------------------------------------------------------------------------------
# Test ThisEvent
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django_bs_test import TestCase
from django.contrib.auth.models import User
from wagtail.core.models import Page
from ls.joyous.models import SpecificCalendarPage
from ls.joyous.models import SimpleEventPage, ThisEvent

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = SpecificCalendarPage(owner = self.user,
                                             slug  = "events",
                                             title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.page = SimpleEventPage(owner = self.user,
                                    slug   = "agfest",
                                    title  = "AgFest",
                                    date   = dt.date(2015,6,5),
                                    time_from = dt.time(11),
                                    time_to   = dt.time(17,30))
        self.calendar.add_child(instance=self.page)
        self.page.save_revision().publish()

    def testInit(self):
        event = ThisEvent(self.page, title="Mud mud mud")
        self.assertEqual(event.page, self.page)
        self.assertEqual(event.title, "Mud mud mud")
        self.assertEqual(event.url, "/events/agfest/")

    def testInit2Args(self):
        with self.assertRaises(TypeError):
            event = ThisEvent(self.page, "Mud mud mud")

    def testRepr(self):
        event = ThisEvent(self.page, url="/events/muddy/")
        self.assertEqual(repr(event),
                         "ThisEvent (title='AgFest', "
                         "page=<SimpleEventPage: AgFest>, "
                         "url='/events/muddy/')")

    def testPageAttrs(self):
        event = ThisEvent(self.page)
        self.assertEqual(event.title, "AgFest")
        self.assertEqual(event.slug, "agfest")
        self.assertEqual(event.date, dt.date(2015,6,5))

    def testTheseAttrs(self):
        event = ThisEvent(self.page)
        event.title = "Mud"
        event.url = "/events/muddy/"
        self.assertEqual(event.title, "Mud")
        self.assertEqual(self.page.title, "AgFest")
        self.assertEqual(event.url, "/events/muddy/")
        self.assertEqual(self.page.url, "/events/agfest/")

    def testNewAttrs(self):
        event = ThisEvent(self.page, foo="foo")
        event.bar = "bar"
        self.assertTrue(hasattr(event, "foo"))
        self.assertFalse(hasattr(self.page, "foo"))
        self.assertEqual(event.foo, "foo")
        self.assertTrue(hasattr(event, "bar"))
        self.assertFalse(hasattr(self.page, "bar"))
        self.assertEqual(event.bar, "bar")

# ------------------------------------------------------------------------------
class TestBackCompat(TestCase):
    """
    ThisEvent maintains backwards compatibility with the namedtuple it used to be
    """
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = SpecificCalendarPage(owner = self.user,
                                             slug  = "events",
                                             title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.page = SimpleEventPage(owner = self.user,
                                    slug   = "agfest",
                                    title  = "AgFest",
                                    date   = dt.date(2015,6,5),
                                    time_from = dt.time(11),
                                    time_to   = dt.time(17,30))
        self.calendar.add_child(instance=self.page)
        self.page.save_revision().publish()

    def testNamedAttrs(self):
        event = ThisEvent("Mud mud mud", self.page, "/events/muddy/")
        self.assertEqual(event.title, "Mud mud mud")
        self.assertEqual(event.page, self.page)
        self.assertEqual(event.url, "/events/muddy/")

    def testExplode(self):
        event = ThisEvent("Mud mud mud", self.page, "/events/muddy/")
        title, page, url = event
        self.assertEqual(title, "Mud mud mud")
        self.assertEqual(page, self.page)
        self.assertEqual(url, "/events/muddy/")

    def testItems(self):
        event = ThisEvent("Mud mud mud", self.page, "/events/muddy/")
        self.assertEqual(len(event), 3)
        self.assertEqual(event[0], "Mud mud mud")
        self.assertEqual(event[1], self.page)
        self.assertEqual(event[2], "/events/muddy/")

    def testSlice(self):
        event = ThisEvent("Mud mud mud", self.page, "/events/muddy/")
        title, page = event[:2]
        self.assertEqual(title, "Mud mud mud")
        self.assertEqual(page, self.page)

    def testAsDict(self):
        event = ThisEvent("Mud mud mud", self.page, "/events/muddy/")
        attrs = event._asdict()
        self.assertEqual(attrs['title'], "Mud mud mud")
        self.assertEqual(attrs['page'], self.page)
        self.assertEqual(attrs['url'], "/events/muddy/")

    def testQuery(self):
        events = list(SimpleEventPage.events.this())
        self.assertEqual(len(events), 1)
        event = events[0]
        self.assertTrue(isinstance(event, ThisEvent))
        self.assertEqual(event.title, "AgFest")
        self.assertEqual(event.page, self.page)
        self.assertEqual(event.url, "/events/agfest/")

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous events models
# ------------------------------------------------------------------------------
import datetime as dt
from operator import attrgetter
from collections import OrderedDict
from django.conf import settings
from django.core.validators import MinValueValidator, MaxValueValidator
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.db import models
from django.db.models import Q
from django.db.models.query import ModelIterable
from django import forms
from django.forms import widgets
from django.template.response import TemplateResponse
from django.utils import timezone
from django.utils import translation
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext, gettext_noop
from wagtail.core.models import Page
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import (FieldPanel, MultiFieldPanel,
        PageChooserPanel)
from wagtail.images.edit_handlers import ImageChooserPanel
from wagtail.contrib.routable_page.models import RoutablePageMixin, route
from wagtail.search import index
from wagtail.admin.forms import WagtailAdminPageForm
from modelcluster.fields import ParentalKey
from ..utils.manythings import hrJoin
from ..utils.mixins import ProxyPageMixin
from ..utils.telltime import (todayUtc, getAwareDatetime, getLocalDatetime,
        getLocalDateAndTime, getLocalDate, getLocalTime, getLocalTimeAtDate)
from ..utils.telltime import getTimeFrom, getTimeTo
from ..utils.telltime import timeFormat, dateFormat
from ..fields import RecurrenceField
from ..forms import FormDefender, BorgPageForm
from ..edit_handlers import (TZDatePanel, ExceptionDatePanel, TimePanel,
        MapFieldPanel)
from ..holidays import Holidays
from .groups import get_group_model_string
from .event_base import (ThisEvent, EventsByDayList,
        EventManager, EventQuerySet, EventPageForm, EventBase)

# ------------------------------------------------------------------------------
# Helper types and constants
# ------------------------------------------------------------------------------
_1day  = dt.timedelta(days=1)
_2days = dt.timedelta(days=2)

# ------------------------------------------------------------------------------
# Event models
# ------------------------------------------------------------------------------
class EventWithHolidaysManager(EventManager):
    def get_queryset(self):
        return self._queryset_class(self.model).live()

    def __call__(self, request, holidays):
        # a shortcut
        return self.get_queryset().auth(request).hols(holidays)

class EventWithHolidaysQuerySet(EventQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.holidays = None

    def _clone(self):
        qs = super()._clone()
        qs.holidays = self.holidays
        return qs

    def _fetch_all(self):
        super()._fetchResults()
        # make sure we all have the same holidays
        if self.holidays is not None:
            for item in self._result_cache:
                for event in getattr(item, 'all_events', [item]):
                    page = getattr(event, 'page', event)
                    page.holidays = self.holidays
        super()._filterResults()

    def hols(self, holidays):
        qs = self._clone()
        qs.holidays = holidays
        return qs

class RecurringEventQuerySet(EventWithHolidaysQuerySet):
    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page, url=page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

    def byDay(self, fromDate, toDate):
        request  = self.request
        holidays = self.holidays
        class ByDayIterable(ModelIterable):
            def __iter__(self):
                evods = EventsByDayList(fromDate, toDate, holidays)
                for page in super().__iter__():
                    exceptions = self.__getExceptionsFor(page)
                    closedHols = self.__getClosedForHolidays(page)
                    startDelta = dt.timedelta(days=page.num_days + 1)
                    for occurence in page.repeat.between(fromDate - startDelta,
                                                         toDate + _2days,
                                                         inc=True):
                        thisEvent = None
                        exception = exceptions.get(occurence)
                        if exception:
                            if exception.title:
                                thisEvent = exception
                        elif closedHols and closedHols._closedOn(occurence):
                            # ClosedForHolidaysPage still affects the event,
                            # even if the user is not authorized
                            if (closedHols.isAuthorized(request) and
                                closedHols.cancellation_title):
                                thisEvent = ThisEvent(closedHols.cancellation_title,
                                                      closedHols,
                                                      closedHols.get_url(request))
                        else:
                            thisEvent = ThisEvent(page, url=page.get_url(request))
                        if thisEvent:
                            pageFromDate = getLocalDate(occurence,
                                                        page.time_from, page.tz)
                            daysDelta = dt.timedelta(days=page.num_days - 1)
                            pageToDate = getLocalDate(occurence + daysDelta,
                                                      page.time_to, page.tz)
                            evods.add(thisEvent, pageFromDate, pageToDate)
                yield from evods

            def __getExceptionsFor(self, page):
                dateRange = (fromDate - _2days, toDate + _2days)
                exceptions = {}
                for extraInfo in ExtraInfoPage.events(request).child_of(page)\
                                     .filter(except_date__range=dateRange):
                    title = extraInfo.extra_title or page.title
                    exceptDate = extraInfo.except_date
                    exceptions[exceptDate] = ThisEvent(title, extraInfo,
                                                       extraInfo.get_url(request))
                for cancellation in CancellationPage.events.child_of(page)   \
                                     .filter(except_date__range=dateRange):
                    # The cancellation still affects the event, even if the
                    # user is not authorized to view the cancellation.
                    if cancellation.isAuthorized(request):
                        title = cancellation.cancellation_title
                        url = cancellation.getCancellationUrl(request)
                    else:
                        title = None
                        url = None
                    exceptDate = cancellation.except_date
                    exceptions[exceptDate] = ThisEvent(title, cancellation, url)
                # TODO consider storing extended cancellations in an interval tree?
                for shutdown in ExtCancellationPage.events.child_of(page)    \
                           .filter(cancelled_from_date__lte=dateRange[1])    \
                           .filter(Q(cancelled_to_date__gte=dateRange[0]) |
                                   Q(cancelled_to_date__isnull = True)):
                    if shutdown.isAuthorized(request):
                        title = shutdown.cancellation_title
                        url = shutdown.get_url(request)
                    else:
                        title = None
                        url = None
                    thisEvent = ThisEvent(title, shutdown, url)
                    for myDate in shutdown._getMyRawDates(*dateRange):
                        exceptions[myDate] = thisEvent
                return exceptions

            def __getClosedForHolidays(self, page):
                closedHols = ClosedForHolidaysPage.events.hols(holidays)     \
                                                  .child_of(page).first()
                return closedHols

        qs = self._clone()
        qs._iterable_class = ByDayIterable
        return qs

# Panel trickery needed as editing proxy models doesn't work yet :-(
class HiddenNumDaysPanel(FieldPanel):
    class widget(widgets.NumberInput):
        def value_from_datadict(self, data, files, name):
            # validation doesn't like num_days disappearing
            return data.get(name, "1")

    def __init__(self, field_name="num_days", *args, **kwargs):
        super().__init__(field_name, *args, **kwargs)

    def render_as_object(self):
        return super().render_as_object() if self._show() else ""

    def render_as_field(self):
        return super().render_as_field() if self._show() else ""

    def _show(self):
        page = getattr(self, 'instance', None)
        if isinstance(page, (MultidayRecurringEventPage,
                             RescheduleMultidayEventPage)):
            retval = True
        else:
            numDays = getattr(page, 'num_days', 0)
            retval = numDays > 1
        return retval

class RecurringEventPageForm(EventPageForm):
    def _checkStartBeforeEnd(self, cleaned_data):
        numDays = cleaned_data.get('num_days', 1)
        if numDays == 1:
            super()._checkStartBeforeEnd(cleaned_data)

class RecurringEventPage(EventBase, Page, metaclass=FormDefender):
    events = EventWithHolidaysManager.from_queryset(RecurringEventQuerySet)()

    class Meta:
        verbose_name = _("recurring event page")
        verbose_name_plural = _("recurring event pages")
        default_manager_name = "objects"

    parent_page_types = ["joyous.CalendarPage",
                         "joyous.SpecificCalendarPage",
                         "joyous.GeneralCalendarPage",
                         get_group_model_string()]
    subpage_types = ['joyous.ExtraInfoPage',
                     'joyous.CancellationPage',
                     'joyous.ExtCancellationPage',
                     'joyous.ClosedForHolidaysPage',
                     'joyous.PostponementPage']
    base_form_class = RecurringEventPageForm
    MAX_REPEAT_COUNT = 500

    repeat   = RecurrenceField(_("repeat"))
    num_days = models.IntegerField(_("number of days"), default=1,
                                   validators=[MinValueValidator(1),
                                               MaxValueValidator(99)])

    content_panels0 = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('repeat')]
    content_panels1 = [
        TimePanel('time_from'),
        TimePanel('time_to'),
        FieldPanel('tz'),
        ] + EventBase.content_panels1
    content_panels = content_panels0 + [HiddenNumDaysPanel()] + content_panels1

    # Anything inheriting from models.Model needs its own __init__ or
    # modeltranslation patch_constructor may break it
    def __init__(self, *args, **kwargs):
        self.holidays = kwargs.pop("holidays", None)
        super().__init__(*args, **kwargs)

    @property
    def next_date(self):
        """
        Date when this event is next scheduled to occur in the local time zone
        (Does not include postponements, but does exclude cancellations)
        """
        myNextDt = self.__after(timezone.localtime(timezone=self.tz))
        if myNextDt is not None:
            return getLocalDate(myNextDt.date(), self.time_from,
                                self.tz, dt.time.min)

    @property
    def _current_datetime_from(self):
        """
        The datetime this event will start or did start in the local timezone, or
        None if it is finished.
        """
        myNow = timezone.localtime(timezone=self.tz)
        timeFrom = getTimeFrom(self.time_from)
        timeTo = getTimeTo(self.time_to)
        # Yes this ignores DST transitions and milliseconds
        timeDelta = dt.timedelta(days    = self.num_days - 1,
                                 hours   = timeTo.hour   - timeFrom.hour,
                                 minutes = timeTo.minute - timeFrom.minute,
                                 seconds = timeTo.second - timeFrom.second)
        myNextDt = self.__after(myNow - timeDelta,
                              excludeCancellations=True,
                              excludeExtraInfo=True)
        if myNextDt is not None:
            return getLocalDatetime(myNextDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _future_datetime_from(self):
        """
        The datetime this event next starts in the local timezone, or None if
        in the past.
        """
        myNextDt = self.__after(timezone.localtime(timezone=self.tz),
                                excludeCancellations=True,
                                excludeExtraInfo=True)
        if myNextDt is not None:
            return getLocalDatetime(myNextDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def prev_date(self):
        """
        Date when this event last occurred in the local time zone
        (Does not include postponements, but does exclude cancellations)
        """
        myPrevDt = self.__before(timezone.localtime(timezone=self.tz))
        if myPrevDt is not None:
            return getLocalDate(myPrevDt.date(), self.time_from,
                                self.tz, dt.time.min)

    @property
    def _past_datetime_from(self):
        """
        The datetime this event previously started in the local time zone, or
        None if it never did.
        """
        myPrevDt = self.__before(timezone.localtime(timezone=self.tz),
                                 excludeCancellations=True,
                                 excludeExtraInfo=True)
        # Exclude extra info pages as this is used for sorting and the extra
        # info pages sit alongside
        if myPrevDt is not None:
            return getLocalDatetime(myPrevDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _first_datetime_from(self):
        """
        The datetime this event first started in the local time zone, or None if
        it never did.
        """
        myFromDt = self._getMyFirstDatetimeFrom()
        localTZ = timezone.get_current_timezone()
        return myFromDt.astimezone(localTZ)

    @property
    def status(self):
        """
        The current status of the event (started, finished or pending).
        """
        myNow = timezone.localtime(timezone=self.tz)
        myDaysDelta = dt.timedelta(days=self.num_days - 1)
        # NB: postponements can be created after the until date
        #     so ignore that
        todayStart = getAwareDatetime(myNow.date(), dt.time.min, self.tz)
        eventStart, event = self.__afterOrPostponedTo(todayStart - myDaysDelta)
        # Will miss any postponements that started more than myDaysDelta ago
        # which might still be active if their num_days is longer than that.
        if eventStart is None:
            return "finished"
        eventDaysDelta = dt.timedelta(days=event.num_days - 1)
        eventFinish = getAwareDatetime(eventStart.date() + eventDaysDelta,
                                       event.time_to, self.tz)
        if event.time_from is None:
            eventStart += _1day
        if eventStart < myNow < eventFinish:
            # if there are two occurences on the same day then we may miss
            # that one of them has started
            return "started"
        if (self.repeat.until and eventFinish < myNow and
            self.__afterOrPostponedTo(myNow)[0] is None):
            # only just wound up, the last occurence was earlier today
            return "finished"

    @property
    def status_text(self):
        """
        A text description of the current status of the event.
        """
        status = self.status
        if status == "finished":
            return _("These events have finished.")
        else:
            return super().status_text

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        offset   = 0
        timeFrom = dateFrom = timeTo = dateTo = None
        myFromDt = self.__getMyFromDt()
        if myFromDt is not None:
            offset = timezone.localtime(myFromDt).toordinal() - myFromDt.toordinal()
            dateFrom, timeFrom = getLocalDateAndTime(myFromDt.date(), self.time_from,
                                                     self.tz, dt.time.min)
            daysDelta = dt.timedelta(days=self.num_days - 1)
            dateTo, timeTo = getLocalDateAndTime(myFromDt.date() + daysDelta,
                                                 self.time_to, self.tz)
        if dateFrom == dateTo:
            retval = _("{repeat} {atTime}").format(
                    repeat=self.repeat._getWhen(offset),
                    atTime=timeFormat(timeFrom, timeTo, gettext("at ")))
        else:
            localNumDays = (dateTo - dateFrom).days + 1
            retval = _("{repeat} {startFinishTime}").format(
                    repeat=self.repeat._getWhen(offset, localNumDays),
                    startFinishTime=timeFormat(timeFrom, timeTo,
                                               prefix=gettext("starting at "),
                                               infix=gettext("finishing at")))
        return retval.strip()

    def _getFromTime(self, atDate=None):
        """
        What was the time of this event?  Due to time zones that depends what
        day we are talking about.  If no day is given, assume today.

        :param atDate: day in local timezone for which we want the time_from
        :rtype: time_from in local timezone
        """
        if atDate is None:
            atDate = timezone.localdate()
        return getLocalTimeAtDate(atDate, self.time_from, self.tz)

    def _futureExceptions(self, request):
        """
        Returns all future extra info, cancellations and postponements created
        for this recurring event
        """
        retval = []
        for extraInfo in ExtraInfoPage.events(request).child_of(self)         \
                                      .future():
            retval.append(extraInfo)
        for cancellation in CancellationPage.events(request).child_of(self)   \
                                            .future():
            postponement = getattr(cancellation, "postponementpage", None)
            if postponement:
                retval.append(postponement)
            else:
                retval.append(cancellation)
        for shutdown in ExtCancellationPage.events(request).child_of(self)   \
                                           .future():
            retval.append(shutdown)
        closedHols = ClosedForHolidaysPage.events(request, self.holidays)    \
                                          .child_of(self).first()
        if closedHols is not None:
            retval.append(closedHols)
        retval.sort(key=attrgetter('_future_datetime_from'))
        # notice these are events not ThisEvents
        return retval

    def _nextOn(self, request):
        """
        Formatted date/time of when this event (including any postponements)
        will next be on
        """
        retval = None
        myNow = timezone.localtime(timezone=self.tz)
        myNextDt, event = self.__afterOrPostponedTo(myNow)
        if myNextDt is not None:
            nextDate, nextTime = getLocalDateAndTime(myNextDt.date(), event.time_from,
                                                     self.tz, dt.time.min)
            retval = "{} {}".format(dateFormat(nextDate),
                                    timeFormat(nextTime, prefix=gettext("at ")))
            if event is not self and event.isAuthorized(request):
                retval = format_html('<a class="inline-link" href="{}">{}</a>',
                                     event.url, retval)
        return retval

    def _occursOn(self, myDate):
        """
        Returns true iff an occurence of this event starts on this date
        (given in the event's own timezone).

        (Does not include postponements, but does exclude cancellations.)
        """
        if myDate not in self.repeat:
            return False
        if CancellationPage.events.child_of(self)                            \
                           .filter(except_date=myDate).exists():
            return False
        if ExtCancellationPage.events.child_of(self)                         \
                           .filter(cancelled_from_date__lte=myDate)          \
                           .filter(Q(cancelled_to_date__gte=myDate) |
                                   Q(cancelled_to_date__isnull = True))      \
                           .exists():
            return False
        closedHols = ClosedForHolidaysPage.events.hols(self.holidays)        \
                                          .child_of(self).first()
        if closedHols and closedHols._closedOn(myDate):
            return False
        return True

    def _getMyFirstDatetimeFrom(self):
        """
        The datetime this event first started, or None if it never did.
        """
        myStartDt = getAwareDatetime(self.repeat.dtstart, None,
                                     self.tz, dt.time.min)
        return self.__after(myStartDt, excludeCancellations=False)

    def _getMyFirstDatetimeTo(self, myFirstDt=False):
        """
        The datetime this event first finished, or None if it never did.
        """
        if myFirstDt is False:
            myFirstDt = self._getMyFirstDatetimeFrom()
        if myFirstDt is not None:
            daysDelta = dt.timedelta(days=self.num_days - 1)
            return getAwareDatetime(myFirstDt.date() + daysDelta,
                                    self.time_to,
                                    self.tz, dt.time.max)

    def _getMyNextDate(self):
        """
        Date when this event is next scheduled to occur in my time zone
        (Does not include postponements, but does exclude cancellations)
        """
        myNextDt = self.__after(timezone.localtime(timezone=self.tz))
        if myNextDt is not None:
            return myNextDt.date()

    def __getMyFromDt(self):
        """
        Get the datetime of the next event after or before now in my timezone.
        """
        myNow = timezone.localtime(timezone=self.tz)
        return self.__after(myNow) or self.__before(myNow)

    def __afterOrPostponedTo(self, fromDt):
        after = self.__after(fromDt)
        # We know all postponement exception dates are in the parent time zone
        if after:
            # is there a postponed event before that?
            # nb: range is inclusive
            dateRange = (fromDt.date(), after.date())
            postponements = PostponementPage.events.child_of(self)           \
                                     .filter(date__range=(dateRange))        \
                                     .order_by('date', 'time_from')
            for postponement in postponements:
                postDt = getAwareDatetime(postponement.date,
                                          postponement.time_from,
                                          self.tz, dt.time.min)
                postDtMax = getAwareDatetime(postponement.date,
                                             postponement.time_from,
                                             self.tz, dt.time.max)
                if postDt < after and postDtMax >= fromDt:
                    return (postDt, postponement)
        else:
            # is there a postponed event then?
            postponements = PostponementPage.events.child_of(self)           \
                                     .filter(date__gte=fromDt.date())        \
                                     .order_by('date', 'time_from')
            for postponement in postponements:
                postDt = getAwareDatetime(postponement.date,
                                          postponement.time_from,
                                          self.tz, dt.time.min)
                postDtMax = getAwareDatetime(postponement.date,
                                             postponement.time_from,
                                             self.tz, dt.time.max)
                if postDtMax >= fromDt:
                    return (postDt, postponement)

        if after is not None:
            return (after, self)
        else:
            return (None, None)

    def __after(self, fromDt, excludeCancellations=True, excludeExtraInfo=False):
        fromDate = fromDt.date()
        if self.time_from and self.time_from < fromDt.time():
            fromDate += _1day
        exceptions = set()
        shutdowns  = []
        closedHols = None
        if excludeCancellations:
            for cancelled in CancellationPage.events.child_of(self)          \
                           .filter(except_date__gte=fromDate):
                exceptions.add(cancelled.except_date)
            # TODO consider storing extended cancellations in an interval tree?
            shutdowns = ExtCancellationPage.events.child_of(self) \
                           .filter(Q(cancelled_to_date__gte=fromDate) |
                                   Q(cancelled_to_date__isnull = True))
            closedHols = ClosedForHolidaysPage.events.hols(self.holidays)    \
                                              .child_of(self).first()
        if excludeExtraInfo:
            for info in ExtraInfoPage.events.child_of(self)                  \
                                     .filter(except_date__gte=fromDate)      \
                                     .exclude(extra_title=""):
                exceptions.add(info.except_date)
        for occurence in self.repeat.xafter(fromDate,
                                            count=self.MAX_REPEAT_COUNT,
                                            inc=True):
            if occurence in exceptions:
                continue
            shutdown = next((shutdown for shutdown in shutdowns
                             if shutdown._closedOn(occurence)), None)
            if shutdown is not None:
                if shutdown.cancelled_to_date is None:
                    break
                else:
                    # the next line is run; coverage is wrong
                    continue  # pragma: no cover
            if closedHols and closedHols._closedOn(occurence):
                continue
            return getAwareDatetime(occurence, self.time_from,
                                    self.tz, dt.time.min)

    def __before(self, fromDt, excludeCancellations=True, excludeExtraInfo=False):
        fromDate = fromDt.date()
        if self.time_from and self.time_from > fromDt.time():
            fromDate -= _1day
        exceptions = set()
        shutdowns  = []
        if excludeCancellations:
            for cancelled in CancellationPage.events.child_of(self)          \
                           .filter(except_date__lte=fromDate):
                exceptions.add(cancelled.except_date)
            shutdowns = ExtCancellationPage.events.child_of(self)            \
                           .filter(cancelled_from_date__lte=fromDate)
            closedHols = ClosedForHolidaysPage.events.hols(self.holidays)    \
                                              .child_of(self).first()
        if excludeExtraInfo:
            for info in ExtraInfoPage.events.child_of(self)                  \
                                     .filter(except_date__lte=fromDate)      \
                                     .exclude(extra_title=""):
                exceptions.add(info.except_date)
        last = None
        for occurence in self.repeat:
            if occurence > fromDate:
                break
            if occurence in exceptions:
                continue
            if any(shutdown._closedOn(occurence) for shutdown in shutdowns):
                continue
            if closedHols and closedHols._closedOn(occurence):
                continue
            last = occurence

        if last is not None:
            return getAwareDatetime(last, self.time_from, self.tz, dt.time.min)

# ------------------------------------------------------------------------------
class MultidayRecurringEventPage(ProxyPageMixin, RecurringEventPage):
    """
    A proxy of RecurringEventPage that exposes the hidden num_days field.
    """
    class Meta(ProxyPageMixin.Meta):
        verbose_name = _("multiday recurring event page")
        verbose_name_plural = _("multiday recurring event pages")

    subpage_types = ['joyous.ExtraInfoPage',
                     'joyous.CancellationPage',
                     'joyous.ExtCancellationPage',
                     'joyous.RescheduleMultidayEventPage']
    template = "joyous/recurring_event_page.html"

    content_panels = RecurringEventPage.content_panels0 + [
        FieldPanel('num_days'),
        ] + RecurringEventPage.content_panels1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

# ------------------------------------------------------------------------------
# EventException models
# ------------------------------------------------------------------------------
class EventExceptionBase(models.Model):
    class Meta:
        abstract = True

    # overrides is also the parent, but parent is not set until the
    # child is saved and added.  (NB: is published version of parent)
    overrides = models.ForeignKey('joyous.RecurringEventPage',
                                  null=True, blank=False,
                                  related_name='+',
                                  verbose_name=_("overrides"),
                                  # can't set to CASCADE, so go with SET_NULL
                                  on_delete=models.SET_NULL)
    overrides.help_text = _("The recurring event that we are updating.")

    # Original properties
    num_days    = property(attrgetter("overrides.num_days"))
    time_from   = property(attrgetter("overrides.time_from"))
    time_to     = property(attrgetter("overrides.time_to"))
    tz          = property(attrgetter("overrides.tz"))
    group       = property(attrgetter("overrides.group"))
    category    = property(attrgetter("overrides.category"))
    image       = property(attrgetter("overrides.image"))
    location    = property(attrgetter("overrides.location"))
    website     = property(attrgetter("overrides.website"))

    # NB: exceptions do not need _first_datetime_from

    # Init these variables to prevent template DEBUG messages
    # Yes, this is very ugly.  An alternative solution would be welcome.
    cancellation_details = None
    extra_information    = None
    postponed_from_when  = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def at(self):
        """
        A string describing what time the event starts (in the local time zone).
        """
        return timeFormat(self._getFromTime())

    @property
    def overrides_repeat(self):
        """
        The recurrence rule of the event being overridden.
        """
        return getattr(self.overrides, 'repeat', None)

    def get_context(self, request, *args, **kwargs):
        retval = super().get_context(request, *args, **kwargs)
        retval['overrides'] = self.overrides
        retval['themeCSS'] = getattr(settings, "JOYOUS_THEME_CSS", "")
        return retval

    # TOO CONFUSING.  Stick with English and the event's timezone
    # def get_admin_display_title(self):
    #     if self.has_unpublished_changes:
    #         # WARNING this could be expensive
    #         page = self.get_latest_revision().as_page_object()
    #     else:
    #         page = self
    #     return page.local_title

    def isAuthorized(self, request):
        """
        Is the user authorized for the requested action with this event?
        """
        restrictions = self.get_view_restrictions()
        if restrictions and request is None:
            return False
        else:
            return all(restriction.accept_request(request)
                       for restriction in restrictions)

    def _copyFieldsFromParent(self, parent):
        """
        Copy across field values from the recurring event parent.
        """
        self.overrides = parent

# ------------------------------------------------------------------------------
class DateExceptionQuerySet(EventQuerySet):
    def current(self):
        qs = super().current()
        return qs.filter(except_date__gte = todayUtc() - _1day)

    def future(self):
        qs = super().future()
        return qs.filter(except_date__gte = todayUtc() - _1day)

    def past(self):
        qs = super().past()
        return qs.filter(except_date__lte = todayUtc() + _1day)

class DateExceptionPageForm(BorgPageForm):
    def _checkSlugAvailable(self, cleaned_data, slugName=None):
        if slugName is None:
            slugName = self.instance.slugName
        description = getattr(self, 'description', "a {}".format(slugName))
        exceptDate = cleaned_data.get('except_date', "invalid")
        slug = "{}-{}".format(exceptDate, slugName)
        if not Page._slug_is_available(slug, self.parent_page, self.instance):
            self.add_error('except_date',
                           _("That date already has") + " " + description.lower())

class DateExceptionBase(EventExceptionBase):
    class Meta:
        abstract = True

    events = EventManager.from_queryset(DateExceptionQuerySet)()

    except_date = models.DateField(_("For Date"))
    except_date.help_text = _("For this date")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def local_title(self):
        """
        Localised version of the human-readable title of the page.
        """
        name = self.title.partition(" for ")[0]
        exceptDate = getLocalDate(self.except_date, self.time_from, self.tz)
        title = _("{exception} for {date}").format(exception=_(name),
                                                   date=dateFormat(exceptDate))
        return title

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        return self._getLocalWhen(self.except_date, self.num_days)

    def full_clean(self, *args, **kwargs):
        """
        Apply fixups that need to happen before per-field validation occurs.
        Sets the page's title.
        """
        name = self.slugName.title()
        # generate the title and slug in English
        # the translation of the title will happen in the property local_title
        with translation.override("en"):
            # FIXME? year may be missing which makes the title ambiguous
            self.title = "{} for {}".format(name, dateFormat(self.except_date))
            self.slug = "{}-{}".format(self.except_date, self.slugName)
        super().full_clean(*args, **kwargs)

    def _getLocalWhen(self, date_from, num_days=1):
        """
        Returns a string describing when the event occurs (in the local time zone).
        """
        dateFrom, timeFrom = getLocalDateAndTime(date_from, self.time_from,
                                                 self.tz, dt.time.min)
        if num_days > 1 or self.time_to is not None:
            daysDelta = dt.timedelta(days=self.num_days - 1)
            dateTo, timeTo = getLocalDateAndTime(date_from + daysDelta,
                                                 self.time_to, self.tz)
        else:
            dateTo = dateFrom
            timeTo = None

        if dateFrom == dateTo:
            retval = _("{date} {atTime}").format(
                    date=dateFormat(dateFrom),
                    atTime=timeFormat(timeFrom, timeTo, gettext("at ")))
        else:
            # Friday the 10th of April for 3 days starting at 1pm finishing at 10am
            # TODO Is "Friday the 10th of April at 1pm to Sunday the 12th of April at 10am" better????
            localNumDays = (dateTo - dateFrom).days + 1
            retval = _("{date} for {n} days {startTime}").format(
                    date=dateFormat(dateFrom),
                    n=localNumDays,
                    startTime=timeFormat(timeFrom, prefix=gettext("starting at ")))
            retval = _("{dateAndStartTime} {finishTime}").format(
                    dateAndStartTime=retval.strip(),
                    finishTime=timeFormat(timeTo, prefix=gettext("finishing at ")))
        return retval.strip()

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone).
        """
        return getLocalTime(self.except_date, self.time_from, self.tz)

    def _getFromDt(self):
        """
        Datetime that the event starts (in the local time zone).
        """
        return getLocalDatetime(self.except_date, self.time_from, self.tz)

    def _getToDt(self):
        """
        Datetime that the event ends (in the local time zone).
        """
        daysDelta = dt.timedelta(days=self.num_days - 1)
        return getAwareDatetime(self.except_date + daysDelta, self.time_to, self.tz)

    def _copyFieldsFromParent(self, parent):
        """
        Copy across field values from the recurring event parent.
        """
        super()._copyFieldsFromParent(parent)
        self.except_date = parent._getMyNextDate()

# ------------------------------------------------------------------------------
class ExtraInfoQuerySet(DateExceptionQuerySet):
    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.extra_title, page,
                                    page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

class ExtraInfoPageForm(DateExceptionPageForm):
    description = _("Extra Information")

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        return cleaned_data

class ExtraInfoPage(DateExceptionBase, Page, metaclass=FormDefender):
    class Meta:
        verbose_name = _("extra event information")
        verbose_name_plural = _("extra event information")
        default_manager_name = "objects"

    events = EventManager.from_queryset(ExtraInfoQuerySet)()
    parent_page_types = ["joyous.RecurringEventPage",
                         "joyous.MultidayRecurringEventPage"]
    subpage_types = []
    base_form_class = ExtraInfoPageForm
    slugName    = "extra-info"
    gettext_noop("Extra-Info")

    extra_title = models.CharField(_("title"), max_length=255, blank=True)
    extra_title.help_text = _("A more specific title for this occurence (optional)")
    extra_information = RichTextField(_("extra information"), blank=True)
    extra_information.help_text = _("Information just for this date")

    search_fields = Page.search_fields + [
        index.SearchField('extra_title'),
        index.SearchField('extra_information'),
    ]
    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        FieldPanel('extra_title', classname="full title"),
        FieldPanel('extra_information', classname="full"),
        ]
    promote_panels = []

    # Original properties
    details     = property(attrgetter("overrides.details"))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def status(self):
        """
        The current status of the event (started, finished or pending).
        """
        now = timezone.localtime()
        if self._getToDt() < now:
           return "finished"
        elif self._getFromDt() < now:
           return "started"

    @property
    def status_text(self):
        """
        A text description of the current status of the event.
        """
        status = self.status
        if status == "finished":
            return _("This event has finished.")
        elif status == "started":
            return _("This event has started.")
        else:
            return ""

    @property
    def _current_datetime_from(self):
        """
        The datetime this event will start or did start in the local timezone, or
        None if it is finished.
        """
        # _occursOn means a lookup of CancellationPage and ClosedForHolidaysPage
        # possible performance problem?
        if not self.overrides._occursOn(self.except_date):
            return None
        fromDt = self._getFromDt()
        toDt = self._getToDt()
        return fromDt if toDt >= timezone.localtime() else None

    @property
    def _future_datetime_from(self):
        """
        The datetime this event next starts in the local timezone, or None if
        in the past.
        """
        # possible performance problem?
        if not self.overrides._occursOn(self.except_date):
            return None
        fromDt = self._getFromDt()
        return fromDt if fromDt >= timezone.localtime() else None

    @property
    def _past_datetime_from(self):
        """
        The datetime this event previously started in the local timezone, or
        None if it never did.
        """
        # possible performance problem?
        if not self.overrides._occursOn(self.except_date):
            return None
        fromDt = self._getFromDt()
        return fromDt if fromDt < timezone.localtime() else None

# ------------------------------------------------------------------------------
class CancellationQuerySet(DateExceptionQuerySet):
    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.cancellation_title, page,
                                    page.getCancellationUrl(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

class CancellationPageForm(DateExceptionPageForm):
    description = _("a cancellation")

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        self._checkSlugAvailable(cleaned_data, "postponement")
        return cleaned_data

class CancellationBase(models.Model):
    class Meta:
        abstract = True

    cancellation_title = models.CharField(_("title"), max_length=255, blank=True)
    cancellation_title.help_text = _("Show in place of cancelled event "
                                     "(Leave empty to show nothing)")
    cancellation_details = RichTextField(_("details"), blank=True)
    cancellation_details.help_text = _("Why was the event cancelled?")

    search_fields = Page.search_fields + [
        index.SearchField('cancellation_title'),
        index.SearchField('cancellation_details'),
    ]

    cancellation_panel = MultiFieldPanel([
            FieldPanel('cancellation_title', classname="full title"),
            FieldPanel('cancellation_details', classname="full")],
            heading=_("Cancellation"))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def status(self):
        """
        The current status of the event
        """
        return "cancelled"

    @property
    def status_text(self):
        """
        A text description of the current status of the event.
        """
        return _("This event has been cancelled.")

class CancellationPage(CancellationBase, DateExceptionBase, Page):
    class Meta:
        verbose_name = _("cancellation")
        verbose_name_plural = _("cancellations")
        default_manager_name = "objects"

    events = EventManager.from_queryset(CancellationQuerySet)()
    parent_page_types = ["joyous.RecurringEventPage",
                         "joyous.MultidayRecurringEventPage"]
    subpage_types = []
    base_form_class = CancellationPageForm
    slugName = "cancellation"
    gettext_noop("Cancellation")

    search_fields = Page.search_fields + CancellationBase.search_fields

    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        CancellationBase.cancellation_panel,
        ]
    promote_panels = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _current_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if it would have finished by now.
        """
        if self.except_date not in self.overrides.repeat:
            return None
        fromDt = self._getFromDt()
        toDt = self._getToDt()
        return fromDt if toDt >= timezone.localtime() else None

    @property
    def _future_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if that is in the past.
        """
        if self.except_date not in self.overrides.repeat:
            return None
        fromDt = self._getFromDt()
        return fromDt if fromDt >= timezone.localtime() else None

    @property
    def _past_datetime_from(self):
        """
        The datetime of the event that was cancelled in the local timezone, or
        None if it never would have.
        """
        if self.except_date not in self.overrides.repeat:
            return None
        fromDt = self._getFromDt()
        return fromDt if fromDt < timezone.localtime() else None

    def getCancellationUrl(self, request=None):
        """
        The URL to a page describing the cancellation for Postponements and
        plain Cancellations.
        """
        url = self.get_url(request)
        if hasattr(self, "postponementpage"):
            if url[-1] != '/':
                url += "/from"
            else:
                url += "from/"
        return url

    cancellation_url = property(getCancellationUrl)

# ------------------------------------------------------------------------------
class PostponementQuerySet(EventQuerySet):
    def current(self):
        qs = super().current()
        return qs.filter(date__gte = todayUtc() - _1day)

    def future(self):
        qs = super().future()
        return qs.filter(date__gte = todayUtc() - _1day)

    def past(self):
        qs = super().past()
        return qs.filter(date__lte = todayUtc() + _1day)

    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.postponement_title,
                                    page, page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

    def byDay(self, fromDate, toDate):
        request = self.request
        class ByDayIterable(ModelIterable):
            def __iter__(self):
                evods = EventsByDayList(fromDate, toDate)
                for page in super().__iter__():
                    thisEvent = ThisEvent(page.postponement_title,
                                          page, page.get_url(request))
                    pageFromDate = getLocalDate(page.date,
                                                page.time_from, page.tz)
                    daysDelta = dt.timedelta(days=page.num_days - 1)
                    pageToDate = getLocalDate(page.date + daysDelta,
                                              page.time_to, page.tz)
                    evods.add(thisEvent, pageFromDate, pageToDate)
                yield from evods

        qs = self._clone()
        qs._iterable_class = ByDayIterable
        return qs.filter(date__range=(fromDate - _1day, toDate + _1day))

class PostponementPageForm(DateExceptionPageForm):
    description = _("a postponement")

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        self._checkSlugAvailable(cleaned_data, "cancellation")
        self._checkStartBeforeEnd(cleaned_data)
        return cleaned_data

    def _checkStartBeforeEnd(self, cleaned_data):
        numDays   = cleaned_data.get('num_days', 1)
        startTime = getTimeFrom(cleaned_data.get('time_from'))
        endTime   = getTimeTo(cleaned_data.get('time_to'))
        if numDays == 1 and startTime > endTime:
            self.add_error('time_to', _("Event cannot end before it starts"))

class RescheduleEventBase(EventBase):
    """
    This class exists just so that the class attributes defined here are
    picked up by Django before the properties from EventExceptionBase,
    as well as before EventBase.
    """
    class Meta:
        abstract = True

    # num_days from here to come before overrides.num_days
    num_days = models.IntegerField(_("number of days"), default=1,
                                   validators=[MinValueValidator(1),
                                               MaxValueValidator(99)])
    # Original properties
    tz          = property(attrgetter("overrides.tz"))
    group       = property(attrgetter("overrides.group"))
    uid         = property(attrgetter("overrides.uid"))
    group_page  = None
    get_context = DateExceptionBase.get_context

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class PostponementPage(RoutablePageMixin, RescheduleEventBase, CancellationPage):
    class Meta:
        verbose_name = _("postponement")
        verbose_name_plural = _("postponements")
        default_manager_name = "objects"

    events = EventManager.from_queryset(PostponementQuerySet)()
    parent_page_types = ["joyous.RecurringEventPage"]
    subpage_types = []
    base_form_class = PostponementPageForm
    slugName = "postponement"
    gettext_noop("Postponement")

    postponement_title = models.CharField(_("title"), max_length=255)
    postponement_title.help_text = _("The title for the postponed event")
    date = models.DateField(_("date"))

    search_fields = Page.search_fields + [
        index.SearchField('postponement_title'),
    ]

    postponement_panel0 = [
            FieldPanel('postponement_title', classname="full title"),
            ImageChooserPanel('image'),
            FieldPanel('date')]
    postponement_panel1 = [
            TimePanel('time_from'),
            TimePanel('time_to'),
            FieldPanel('details', classname="full"),
            MapFieldPanel('location'),
            FieldPanel('website')]
    postponement_panel = MultiFieldPanel(
            postponement_panel0 + [HiddenNumDaysPanel()] + postponement_panel1,
            heading=_("Postponed to"))
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        CancellationBase.cancellation_panel,
        postponement_panel,
    ]
    promote_panels = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @route(r"^from/$")
    def serveCancellation(self, request):
        return TemplateResponse(request,
                                "joyous/postponement_page_from.html",
                                self.get_context(request))

    @property
    def when(self):
        """
        A string describing when the postponement occurs (in the local time zone).
        """
        return DateExceptionBase._getLocalWhen(self, self.date, self.num_days)

    @property
    def postponed_from_when(self):
        """
        A string describing when the event was postponed from (in the local time zone).
        """
        what = self.what
        if what:
            return _("{what} from {when}").format(what=what,
                                                  when=self.cancellationpage.when)

    @property
    def what(self):
        """
        May return a 'postponed' or 'rescheduled' string depending what
        the start and finish time of the event has been changed to.
        """
        originalFromDt = dt.datetime.combine(self.except_date,
                                             getTimeFrom(self.overrides.time_from))
        changedFromDt = dt.datetime.combine(self.date, getTimeFrom(self.time_from))
        originalDaysDelta = dt.timedelta(days=self.overrides.num_days - 1)
        originalToDt = dt.datetime.combine(self.except_date + originalDaysDelta,
                                           getTimeTo(self.overrides.time_to))
        changedDaysDelta = dt.timedelta(days=self.num_days - 1)
        changedToDt = dt.datetime.combine(self.except_date + changedDaysDelta,
                                          getTimeTo(self.time_to))
        if originalFromDt < changedFromDt:
            return _("Postponed")
        elif originalFromDt > changedFromDt or originalToDt != changedToDt:
            return _("Rescheduled")
        else:
            return None

    @property
    def postponed_from(self):
        """
        Date that the event was postponed from (in the local time zone).
        """
        fromDate = getLocalDate(self.except_date, self.time_from, self.tz)
        return dateFormat(fromDate)

    @property
    def postponed_to(self):
        """
        Date that the event was postponed to (in the local time zone).
        """
        toDate = getLocalDate(self.date, self.time_from, self.tz)
        return dateFormat(toDate)

    def _getFromTime(self, atDate=None):
        """
        Time that the postponement starts (in the local time zone).
        """
        return getLocalTime(self.date, self.time_from, self.tz)

    def _getFromDt(self):
        """
        Datetime that the postponement starts (in the local time zone).
        """
        return getLocalDatetime(self.date, self.time_from, self.tz)

    def _getToDt(self):
        """
        Datetime that the postponement ends (in the local time zone).
        """
        daysDelta = dt.timedelta(days=self.num_days - 1)
        return getLocalDatetime(self.date + daysDelta, self.time_to, self.tz)

    def _copyFieldsFromParent(self, parent):
        """
        Copy across field values from the recurring event parent.
        """
        super()._copyFieldsFromParent(parent)
        parentFields = set()
        for panel in parent.content_panels:
            parentFields.update(panel.required_fields())
        pageFields = set()
        for panel in self.content_panels:
            pageFields.update(panel.required_fields())
        commonFields = parentFields & pageFields
        for name in commonFields:
            setattr(self, name, getattr(parent, name))
        if self.except_date:
            self.date = self.except_date + dt.timedelta(days=1)
        self.postponement_title = parent.title

# ------------------------------------------------------------------------------
class RescheduleMultidayEventPage(ProxyPageMixin, PostponementPage):
    """
    A proxy of PostponementPage that exposes the hidden num_days field.
    """
    class Meta(ProxyPageMixin.Meta):
        verbose_name = _("postponement")
        verbose_name_plural = _("postponements")

    parent_page_types = ["joyous.MultidayRecurringEventPage"]
    template = "joyous/postponement_page.html"

    postponement_panel = MultiFieldPanel(
            PostponementPage.postponement_panel0 +
            [FieldPanel('num_days')] +
            PostponementPage.postponement_panel1,
            heading=_("Postponed to"))
    content_panels = [
        PageChooserPanel('overrides'),
        ExceptionDatePanel('except_date'),
        PostponementPage.cancellation_panel,
        postponement_panel,
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

# ------------------------------------------------------------------------------
class ClosedFor(models.Model):
    """The holidays we are closed for."""
    class Meta:
        ordering = ['pk']
        unique_together = ['page', 'name']

    page = ParentalKey('joyous.ClosedForHolidaysPage',
                       on_delete=models.CASCADE,
                       related_name="closed_for")
    name = models.CharField(_("name"), max_length=100)
    # Storing holidays by name is a violation of 1NF, but holidays are
    # defined programmatically not in the database and the definitions
    # may change, so there is no holiday_id which can be used instead.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # used by ChoiceWidget.format_value
    def __str__(self):
        return self.name

class ClosedForHolidaysQuerySet(EventWithHolidaysQuerySet):
    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.cancellation_title, page,
                                    page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

class ClosedForHolidaysPageForm(BorgPageForm):
    class Media:
        css = { 'all': ["admin/css/widgets.css",
                        "joyous/css/holidays_admin.css"] }
        js = ["/django-admin/jsi18n",
              "joyous/js/holidays_admin.js"]

    holidays_class = Holidays
    description = _("closed for holidays")
    closed_for = forms.MultipleChoiceField(label=_("These holidays"),
                                           required=False,
                                           widget=FilteredSelectMultiple(
                                                            _("Holidays"),
                                                            is_stacked=False))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['closed_for'].choices = self._holidayChoices()
        self.initial['closed_for'] = list(self.instance.closed_for.all())

    def save(self, commit=True):
        self._saveChosenHolidays()
        page = super().save(commit)
        return page

    def _holidayChoices(self):
        if self.instance.holidays is not None:
            holidays = self.instance.holidays
        else:
            holidays = self.holidays_class()
        retval = [(name, name) for name in holidays.names()]
        return retval

    def _saveChosenHolidays(self):
        chosen = self.cleaned_data.get('closed_for', [])
        initial = {day.name: day for day in self.initial['closed_for']}
        days = [initial.get(name, ClosedFor(name=name)) for name in chosen]
        self.instance.closed_for.set(days)

class ClosedForHolidaysPage(CancellationBase, EventExceptionBase, Page,
                            metaclass=FormDefender):
    class Meta:
        verbose_name = _("closed for holidays")
        verbose_name_plural = _("closed for holidays")
        default_manager_name = "objects"

    events = EventWithHolidaysManager.from_queryset(ClosedForHolidaysQuerySet)()
    parent_page_types = ["joyous.RecurringEventPage"]
    # TODO add "joyous.MultidayRecurringEventPage" : How would that work?
    # A. If the first day of the event is a holiday then the event is cancelled, or
    # B. If a holiday occurs anywhere in the event the event is cancelled, or
    # C. If every day of the event is a holiday then the event is cancelled?
    subpage_types = []
    base_form_class = ClosedForHolidaysPageForm
    MAX_REPEAT_COUNT = RecurringEventPage.MAX_REPEAT_COUNT
    MAX_DATETIME = timezone.make_aware(dt.datetime.max - _2days)
    MIN_DATETIME = timezone.make_aware(dt.datetime.min + _2days)

    all_holidays = models.BooleanField(default=True)
    all_holidays.help_text = "Cancel any occurence of this event on a public holiday"

    search_fields = Page.search_fields + CancellationBase.search_fields
    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        MultiFieldPanel([
            FieldPanel('all_holidays'),
            FieldPanel('closed_for')],
            heading=_("Closed For")),
        CancellationBase.cancellation_panel,
        ]
    promote_panels = []

    @classmethod
    def can_create_at(cls, parent):
        retval = (super().can_create_at(parent) and
                  not cls.objects.descendant_of(parent).exists())
        return retval

    def __init__(self, *args, **kwargs):
        self.holidays = kwargs.pop("holidays", None)
        super().__init__(*args, **kwargs)
        self.__closedSet = None

    def full_clean(self, *args, **kwargs):
        """
        Apply fixups that need to happen before per-field validation occurs.
        Sets the page's title.
        """
        self.title = "Closed for holidays"
        self.slug  = "closed-for-holidays"
        super().full_clean(*args, **kwargs)

    @property
    def local_title(self):
        """
        Localised version of the human-readable title of the page.
        """
        return _("Closed for holidays")

    @property
    def status_text(self):
        """
        A text description of the current status of the event.
        """
        return _("Closed for holidays.")

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        if self.all_holidays:
            retval = _("Closed on all holidays")
        else:
            retval = _("Closed on {}").format(hrJoin(self.closed))
        return retval

    @property
    def closed(self):
        """
        The list of holidays we are closed for, or "ALL" if we are closed for
        all holidays.
        """
        if self.all_holidays:
            retval = "ALL"
        else:
            retval = [day.name for day in self.closed_for.all()]
        return retval

    @property
    def _current_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if it would have finished by now.
        """
        # TODO this code would need to change if this is going to support
        # MultidayRecurringEventPage
        return self._future_datetime_from

    @property
    def _future_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if that is in the past.
        """
        if self.holidays is None:
            return self.MAX_DATETIME
        myNextDt = self.__after(timezone.localtime(timezone=self.tz))
        if myNextDt is not None:
            return getLocalDatetime(myNextDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _past_datetime_from(self):
        """
        The datetime of the event that was cancelled in the local timezone, or
        None if it never would have.
        """
        if self.holidays is None:
            return self.MIN_DATETIME
        myPrevDt = self.__before(timezone.localtime(timezone=self.tz))
        if myPrevDt is not None:
            return getLocalDatetime(myPrevDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _closed_for_dates(self):
        """
        Return all the dates which we are closed for in the local timezone
        """
        for occurence in self._getMyDates():
            yield getLocalDate(occurence, self.time_from, self.tz)

    def _getMyDates(self):
        """
        Return all the dates which we are closed for in the event timezone
        """
        if self.holidays is not None:
            self._cacheClosedSet()
            if not self.__closedSet:
                return None
            n = 0
            for occurence in self.overrides.repeat:
                n += 1
                if n > self.MAX_REPEAT_COUNT:
                    # that's enough, bailing out
                    return None
                if self._closedOn(occurence):
                    n = 0
                    yield occurence

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone).
        """
        return self.overrides._getFromTime(atDate)

    def _cacheClosedSet(self):
        """
        Cache the set of names of the holidays we are closed for.
        This needs to be re-called if the page's fields change.
        """
        closed = self.closed
        if closed == "ALL":
            self.__closedSet = "ALL"
        else:
            self.__closedSet = set(self.closed)

    def _closedOn(self, myDate):
        """
        Are we closed for holidays on myDate?
        """
        if self.holidays is None:
            return False
        if self.__closedSet is None:
            self._cacheClosedSet()
        if not self.__closedSet:
            return False
        holiday = self.holidays.get(myDate)
        if holiday:
            if self.all_holidays:
                return True
            names = holiday.split(", ")
            return any(name in self.__closedSet for name in names)
        return False

    def __after(self, fromDt, excludeCancellations=True):
        fromDate = fromDt.date()
        if self.time_from and self.time_from < fromDt.time():
            fromDate += _1day
        self._cacheClosedSet()
        exceptions = set()
        shutdowns  = []
        if excludeCancellations:
            for cancelled in CancellationPage.events.child_of(self.overrides) \
                                     .filter(except_date__gte=fromDate):
                exceptions.add(cancelled.except_date)
            shutdowns = ExtCancellationPage.events.child_of(self.overrides)  \
                           .filter(Q(cancelled_to_date__gte=fromDate) |
                                   Q(cancelled_to_date__isnull = True))
        repeat = self.overrides.repeat
        for n, occurence in enumerate(repeat.xafter(fromDate, inc=True)):
            if n > self.MAX_REPEAT_COUNT:
                # that's enough, bailing out, return two days before max
                # (so we can still do TZ conversions on the result)
                return self.MAX_DATETIME
            if occurence in exceptions:
                continue
            shutdown = next((shutdown for shutdown in shutdowns
                             if shutdown._closedOn(occurence)), None)
            if shutdown is not None:
                if shutdown.cancelled_to_date is None:
                    break
                else:
                    # the next line is run; coverage is wrong
                    continue  # pragma: no cover
            if not self._closedOn(occurence):
                continue
            return getAwareDatetime(occurence, self.time_from,
                                    self.tz, dt.time.min)

    def __before(self, fromDt, excludeCancellations=True):
        fromDate = fromDt.date()
        if self.time_from and self.time_from > fromDt.time():
            fromDate -= _1day
        self._cacheClosedSet()
        exceptions = set()
        shutdowns  = []
        if excludeCancellations:
            for cancelled in CancellationPage.events.child_of(self.overrides) \
                                     .filter(except_date__lte=fromDate):
                exceptions.add(cancelled.except_date)
            shutdowns = ExtCancellationPage.events.child_of(self.overrides)  \
                           .filter(cancelled_from_date__lte=fromDate)
        last = None
        repeat = self.overrides.repeat
        for occurence in repeat:
            if occurence > fromDate:
                break
            if occurence in exceptions:
                continue
            if any(shutdown._closedOn(occurence) for shutdown in shutdowns):
                continue
            if not self._closedOn(occurence):
                continue
            last = occurence
        if last is not None:
            return getAwareDatetime(last, self.time_from, self.tz, dt.time.min)

# ------------------------------------------------------------------------------
class ExtCancellationQuerySet(EventQuerySet):
    def current(self):
        qs = super().current()
        return qs.filter(Q(cancelled_to_date__gte=todayUtc() - _1day) |
                         Q(cancelled_to_date__isnull=True))

    def future(self):
        qs = super().future()
        return qs.filter(Q(cancelled_to_date__gte=todayUtc() - _1day) |
                         Q(cancelled_to_date__isnull=True))

    def past(self):
        qs = super().past()
        return qs.filter(cancelled_from_date__lte=todayUtc() + _1day)

    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.cancellation_title, page,
                                    page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

class ExtCancellationPageForm(BorgPageForm):
    class Media:
        css = { 'all': ["joyous/css/recurrence_admin.css"] }

    def clean(self):
        cleaned_data = super().clean()
        self._checkSlugAvailable(cleaned_data)
        return cleaned_data

    def _checkSlugAvailable(self, cleaned_data):
        fromDate = cleaned_data.get('cancelled_from_date', "invalid")
        siblings = self.parent_page.get_children().not_page(self.instance)
        pattern = r"{}-.*-cancellation".format(fromDate)
        if siblings.filter(slug__regex=pattern).exists():
            self.add_error('cancelled_from_date',
                           _("There is already an extended cancellation for then"))

class ExtCancellationPage(CancellationBase, EventExceptionBase, Page,
                          metaclass=FormDefender):
    class Meta:
        verbose_name = _("extended cancellation")
        verbose_name_plural = _("extended cancellations")
        default_manager_name = "objects"

    events = EventManager.from_queryset(ExtCancellationQuerySet)()
    parent_page_types = ["joyous.RecurringEventPage",
                         "joyous.MultidayRecurringEventPage"]
    subpage_types = []
    base_form_class = ExtCancellationPageForm
    FAR_DATE = dt.date(3000,12,31)

    cancelled_from_date = models.DateField(_("From Date"))
    cancelled_from_date.help_text = _("Cancelled from this date")
    cancelled_to_date = models.DateField(_("To Date"), null=True, blank=True)
    cancelled_to_date.help_text = _('Cancelled to this date '
                                    '(Leave empty for "until further notice")')

    search_fields = Page.search_fields + CancellationBase.search_fields
    # Note title is not displayed
    content_panels = [
        PageChooserPanel('overrides'),
        TZDatePanel('cancelled_from_date'),
        TZDatePanel('cancelled_to_date'),
        CancellationBase.cancellation_panel,
        ]
    promote_panels = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def local_title(self):
        """
        Localised version of the human-readable title of the page.
        """
        title = _("Cancellation from {fromDate} {untilWhen}").format(
                            fromDate=dateFormat(self.cancelled_from_date),
                            untilWhen=self.until_when)
        return title

    @property
    def until_when(self):
        """
        A string describing how long we are cancelled for
        """
        if self.cancelled_to_date is not None:
            retval = _("to {}").format(dateFormat(self.cancelled_to_date))
        else:
            retval = _("until further notice")
        return retval

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        retval = _("Cancelled from {fromDate} {untilWhen}").format(
                            fromDate=dateFormat(self.cancelled_from_date),
                            untilWhen=self.until_when)
        return retval

    @property
    def _current_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if it would have finished by now.
        """
        myNow = timezone.localtime(timezone=self.tz)
        timeFrom = getTimeFrom(self.time_from)
        timeTo = getTimeTo(self.time_to)
        # Yes this ignores DST transitions and milliseconds
        timeDelta = dt.timedelta(days    = self.num_days - 1,
                                 hours   = timeTo.hour   - timeFrom.hour,
                                 minutes = timeTo.minute - timeFrom.minute,
                                 seconds = timeTo.second - timeFrom.second)
        myNextDt = self.__after(myNow - timeDelta)
        if myNextDt is not None:
            return getLocalDatetime(myNextDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _future_datetime_from(self):
        """
        The datetime the event that was cancelled would start in the local
        timezone, or None if that is in the past.
        """
        myNextDt = self.__after(timezone.localtime(timezone=self.tz))
        if myNextDt is not None:
            return getLocalDatetime(myNextDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    @property
    def _past_datetime_from(self):
        """
        The datetime of the event that was cancelled in the local timezone, or
        None if it never would have.
        """
        myPrevDt = self.__before(timezone.localtime(timezone=self.tz))
        if myPrevDt is not None:
            return getLocalDatetime(myPrevDt.date(), self.time_from,
                                    self.tz, dt.time.max)

    def full_clean(self, *args, **kwargs):
        """
        Apply fixups that need to happen before per-field validation occurs.
        Sets the page's title.
        """
        # generate the title and slug in English
        # the translation of the title will happen in the property local_title
        with translation.override("en"):
            if self.cancelled_to_date is not None:
                # FIXME? year may be missing which makes the title ambiguous
                titleTo = "to " + dateFormat(self.cancelled_to_date)
                slugTo = str(self.cancelled_to_date)
            else:
                titleTo = "until further notice"
                slugTo = ""

            # FIXME? year may be missing which makes the title ambiguous
            self.title = "Cancellation from {} {}".format(
                                dateFormat(self.cancelled_from_date), titleTo)
            self.slug = "{}-{}-cancellation".format(
                                self.cancelled_from_date, slugTo)
        super().full_clean(*args, **kwargs)

    def _getMyDates(self, fromDate=dt.date.min, toDate=FAR_DATE):
        """
        Return all the dates which we are cancelled for in the event timezone.
        Only if the event occurs on those dates.
        Limited by fromDate and toDate if given.
        """
        if self.cancelled_from_date > fromDate:
            fromDate = self.cancelled_from_date
        if (self.cancelled_to_date is not None and
            self.cancelled_to_date < toDate):
            toDate = self.cancelled_to_date
        repeat = self.overrides.repeat
        yield from repeat.between(fromDate, toDate, inc=True)

    def _getMyRawDates(self, fromDate=dt.date.min, toDate=FAR_DATE):
        """
        Return all the dates which we are cancelled for in the event timezone.
        Without considering if the event occurs on those dates.
        Limited by fromDate and toDate if given.
        """
        if self.cancelled_from_date > fromDate:
            fromDate = self.cancelled_from_date
        if (self.cancelled_to_date is not None and
            self.cancelled_to_date < toDate):
            toDate = self.cancelled_to_date
        fromOrd = fromDate.toordinal()
        toOrd   = toDate.toordinal()
        for ord in range(fromOrd, toOrd+1):
            yield dt.date.fromordinal(ord)

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone).
        """
        return self.overrides._getFromTime(atDate)

    def _closedOn(self, myDate):
        """
        Are we closed on myDate?
        """
        if self.cancelled_to_date is not None:
            retval = (self.cancelled_from_date <= myDate <=
                      self.cancelled_to_date)
        else:
            retval = self.cancelled_from_date <= myDate
        return retval

    def __after(self, fromDt):
        fromDate = fromDt.date()
        if self.time_from and self.time_from < fromDt.time():
            fromDate += _1day
        if fromDate < self.cancelled_from_date:
            fromDate = self.cancelled_from_date
        occurence = self.overrides.repeat.after(fromDate, inc=True)
        if (self.cancelled_to_date is None or
            occurence <= self.cancelled_to_date):
            return getAwareDatetime(occurence, self.time_from,
                                    self.tz, dt.time.min)

    def __before(self, fromDt):
        fromDate = fromDt.date()
        if self.time_from and self.time_from > fromDt.time():
            fromDate -= _1day
        if (self.cancelled_to_date is not None and
            fromDate > self.cancelled_to_date):
            fromDate = self.cancelled_to_date
        occurence = self.overrides.repeat.before(fromDate, inc=True)
        if occurence >= self.cancelled_from_date:
            return getAwareDatetime(occurence, self.time_from,
                                    self.tz, dt.time.min)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous Forms
# ------------------------------------------------------------------------------
import warnings
from django.conf import settings
from wagtail.core.models import PageBase
from wagtail.admin.forms import WagtailAdminPageForm

# ------------------------------------------------------------------------------

class BorgPageForm(WagtailAdminPageForm):
    """
    Your page form will be assimilated.
    """
    @classmethod
    def assimilate(cls, form_class):
        if form_class is None or issubclass(form_class, WagtailAdminPageForm):
            cls.assimilated_class = form_class

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if (not getattr(self, 'unimatrix_zero', False) and
            hasattr(self, 'assimilated_class')):
            borg_form_class = self._get_assimilated_form()
            self.assimilated = borg_form_class(*args, **kwargs)
        else:
            self.assimilated = None

    def clean(self):
        if self.assimilated:
            self.assimilated.cleaned_data = self.cleaned_data
            self.assimilated._errors = self._errors
            cleaned_data = self.assimilated.clean()
            self._errors.update(self.assimilated._errors)
            return cleaned_data
        else:
            return super().clean()

    def save(self, commit=True):
        if self.assimilated:
            page = self.assimilated.save(commit)
        else:
            page = super().save(commit)
        return page

    def _get_assimilated_form(self):
        # based on wagtail.admin.edit_handlers.get_form_for_model
        self.assimilated_class.declared_fields.update(self.declared_fields)
        attrs = {'model':    self._meta.model,
                 'fields':   self._meta.fields,
                 'formsets': self._meta.formsets,
                 'widgets':  self._meta.widgets}
        class_name = "Assimilated" + self.assimilated_class.__name__
        bases = (object,)
        if hasattr(self.assimilated_class, 'Meta'):
            bases = (self.assimilated_class.Meta,) + bases
        form_class_attrs = {
            'Meta': type(str('Meta'), bases, attrs),
            'unimatrix_zero': True
        }
        metaclass = type(self.assimilated_class)
        return metaclass(class_name,
                         (self.assimilated_class,),
                         form_class_attrs)

# ------------------------------------------------------------------------------
def _getName(thing):
    return getattr(thing, '__name__', repr(thing))

class FormClassOverwriteWarning(RuntimeWarning):
    pass

class FormDefender(PageBase):
    """
    Metaclass for pages who don't want their base_form_class changed
    """
    def __new__(mcs, name, bases, attrs):
        new_attrs = dict((k if k != 'base_form_class' else '_'+k, v)
                         for k, v in attrs.items())
        cls = super().__new__(mcs, name, bases, new_attrs)
        return cls

    @property
    def base_form_class(cls):
        return cls._base_form_class

    @base_form_class.setter
    def base_form_class(cls, form_class):
        # Probably need to also call get_edit_handler.cache_clear()
        # after changing base_form_class for it to affect the
        # edit_handler.
        my_form_class = cls._base_form_class
        if my_form_class is None:
            cls._base_form_class = form_class
            return

        if getattr(settings, "JOYOUS_DEFEND_FORMS", False):
            if issubclass(my_form_class, BorgPageForm):
                my_form_class.assimilate(form_class)
        else:
            cls._base_form_class = form_class

            # don't generate warning for subclasses
            if not (isinstance(form_class, type) and
                    issubclass(form_class, my_form_class)):
                warning = FormClassOverwriteWarning(
                              "{} has been overwritten with {}, "
                              "consider enabling JOYOUS_DEFEND_FORMS"
                              .format(_getName(my_form_class),
                                      _getName(form_class)))
                warnings.warn(warning, stacklevel=2)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# Generated by Django 2.2.9 on 2020-07-03 01:03

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields
import wagtail.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0041_group_collection_permissions_verbose_name_plural'),
        ('joyous', '0015_auto_20190409_0645'),
    ]

    operations = [
        migrations.CreateModel(
            name='ClosedForHolidaysPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('all_holidays', models.BooleanField(default=True, help_text='Cancel any occurence of this event on a public holiday')),
                ('cancellation_title', models.CharField(blank=True, help_text='Show in place of cancelled event (Leave empty to show nothing)', max_length=255, verbose_name='title')),
                ('cancellation_details', wagtail.core.fields.RichTextField(blank=True, help_text='Why was the event cancelled?', verbose_name='details')),
                ('overrides', models.ForeignKey(help_text='The recurring event that we are updating.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.RecurringEventPage', verbose_name='overrides')),
            ],
            options={
                'verbose_name_plural': 'closed for holidays',
                'verbose_name': 'closed for holidays',
                'default_manager_name': 'objects',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='ClosedFor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('page', modelcluster.fields.ParentalKey(on_delete=django.db.models.deletion.CASCADE, related_name='closed_for', to='joyous.ClosedForHolidaysPage')),
            ],
            options={
                'unique_together': {('page', 'name')},
                'ordering': ['pk'],
            },
        ),
    ]



# Event Base
from .event_base import EventCategory
from .event_base import EventBase
from .event_base import ThisEvent
from .event_base import EventsOnDay

# One Off Events
from .one_off_events import SimpleEventPage
from .one_off_events import MultidayEventPage

# Recurring Events
from .recurring_events import RecurringEventPage
from .recurring_events import MultidayRecurringEventPage
from .recurring_events import EventExceptionBase
from .recurring_events import DateExceptionBase
from .recurring_events import ExtraInfoPage
from .recurring_events import CancellationBase
from .recurring_events import CancellationPage
from .recurring_events import RescheduleEventBase
from .recurring_events import PostponementPage
from .recurring_events import RescheduleMultidayEventPage
from .recurring_events import ExtCancellationPage
from .recurring_events import ClosedForHolidaysPage
from .recurring_events import ClosedFor

# Events API
from .events_api import getAllEventsByDay
from .events_api import getAllEventsByWeek
from .events_api import getAllUpcomingEvents
from .events_api import getAllPastEvents
from .events_api import getGroupUpcomingEvents
from .events_api import getEventFromUid
from .events_api import getAllEvents
from .events_api import removeContentPanels

# Calendars
from .calendar import CalendarPage
from .calendar import CalendarPageForm
from .calendar import SpecificCalendarPage
from .calendar import GeneralCalendarPage

# Groups
from .groups import GroupPage
from .groups import get_group_model
from .groups import get_group_model_string



# ------------------------------------------------------------------------------
# Joyous events models
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from collections import OrderedDict
from uuid import uuid4
from django.conf import settings
from django.db import models
from django.db.models import Q
from django.db.models.query import ModelIterable
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from timezone_field import TimeZoneField
from wagtail.core.query import PageQuerySet
from wagtail.core.models import Page, PageManager, PageViewRestriction
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import (FieldPanel,
        PageChooserPanel, BaseCompositeEditHandler)
from wagtail.images import get_image_model_string
from wagtail.search import index
from wagtail.admin.forms import WagtailAdminPageForm
from ..utils.telltime import getLocalDateAndTime
from ..utils.telltime import getTimeFrom, getTimeTo
from ..utils.telltime import timeFormat, dateFormat
from ..edit_handlers import MapFieldPanel
from ..forms import BorgPageForm
from .groups import get_group_model_string, get_group_model


# ------------------------------------------------------------------------------
# Private
# ------------------------------------------------------------------------------
def _filterContentPanels(panels, remove):
    retval = []
    for panel in panels:
        if isinstance(panel, FieldPanel) and panel.field_name in remove:
            continue
        elif isinstance(panel, BaseCompositeEditHandler):
            panel.children = _filterContentPanels(panel.children, remove)
        retval.append(panel)
    return retval

# ------------------------------------------------------------------------------
# Helper types and constants
# ------------------------------------------------------------------------------
class ThisEvent:
    _fields = ('title', 'page', 'url')

    def __init__(self, *args, **kwargs):
        super().__init__()
        argCount = len(args)
        if argCount == 1:
            # new ctor
            self.page = args[0]
        elif argCount == 3:
            # old ctor
            self.title, self.page, self.url = args
        else:
            raise TypeError("Expected 1 or 3 args not {}".format(argCount))
        for kw, arg in kwargs.items():
            setattr(self, kw, arg)

    def __getattr__(self, attr):
        return getattr(self.page, attr)

    def _asdict(self):
        return OrderedDict(zip(self._fields, self))

    def _astuple(self):
        return (self.title, self.page, self.url)

    def __repr__(self):
        return "ThisEvent (title=%r, page=%r, url=%r)" % self._astuple()

    def __len__(self):
        return 3

    def __getitem__(self, key):
        return self._astuple()[key]

class EventsOnDay:
    """
    The events that occur on a certain day.  Both events that start on that day
    and events that are still continuing.
    """
    def __init__(self, date, holiday=None,
                 days_events=None, continuing_events=None):
        if days_events is None:
            days_events = []
        if continuing_events is None:
            continuing_events = []
        self.date = date
        self.holiday = holiday
        self.days_events = days_events
        self.continuing_events = continuing_events

    @property
    def all_events(self):
        """
        All the events that occur on this day,
        ``days_events + continuing_events``.
        """
        return self.days_events + self.continuing_events

    @property
    def preview(self):
        """
        A short description of some of the events on this day
        (limited to 100 characters).
        """
        return ", ".join(event.title for event in self.all_events)[:100]

    @property
    def weekday(self):
        """
        The weekday abbreviation for this days (e.g. "mon").
        """
        return calendar.day_abbr[self.date.weekday()].lower()

class EventsByDayList(list):
    def __init__(self, fromDate, toDate, holidays=None):
        if holidays is None:
            holidays = {}
        self.fromOrd = fromDate.toordinal()
        self.toOrd   = toDate.toordinal()
        days = [dt.date.fromordinal(ord)
                for ord in range(self.fromOrd, self.toOrd+1)]
        super().__init__(EventsOnDay(day, holidays.get(day)) for day in days)

    def add(self, thisEvent, pageFromDate, pageToDate):
        pageFromOrd = pageFromDate.toordinal()
        pageToOrd   = pageToDate.toordinal()
        dayNum = pageFromOrd - self.fromOrd
        if 0 <= dayNum <= self.toOrd - self.fromOrd:
            self[dayNum].days_events.append(thisEvent)

        for pageOrd in range(pageFromOrd + 1, pageToOrd + 1):
            dayNum = pageOrd - self.fromOrd
            if 0 <= dayNum <= self.toOrd - self.fromOrd:
                self[dayNum].continuing_events.append(thisEvent)

# ------------------------------------------------------------------------------
# Event models
# ------------------------------------------------------------------------------
class EventCategory(models.Model):
    """The category type of an event."""
    class Meta:
        ordering = ["name"]
        verbose_name = _("event category")
        verbose_name_plural = _("event categories")

    code = models.CharField(_("code"), max_length=4, unique=True)
    name = models.CharField(_("name"), max_length=80)

    # Anything inheriting from models.Model needs its own __init__ or
    # modeltranslation patch_constructor may break it
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

# ------------------------------------------------------------------------------
class EventManager(PageManager):
    def get_queryset(self):
        return self._queryset_class(self.model).live()

    def __call__(self, request):
        # a shortcut
        return self.get_queryset().auth(request)

class EventQuerySet(PageQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request    = None
        self.postFilter = None

    def _clone(self):
        qs = super()._clone()
        qs.request    = self.request
        qs.postFilter = self.postFilter
        return qs

    def _fetch_all(self):
        self._fetchResults()
        self._filterResults()

    def _fetchResults(self):
        super()._fetch_all()

    def _filterResults(self):
        if self.postFilter:
            self._result_cache[:] = filter(self.postFilter, self._result_cache)

    def count(self):
        if self.postFilter and self._result_cache is None:
            # if we have a postFilter then force a call to _fetch_all
            self._fetch_all()
        return super().count()

    def upcoming(self):
        if getattr(settings, "JOYOUS_UPCOMING_INCLUDES_STARTED", False):
            return self.current()
        else:
            return self.future()

    def current(self):
        qs = self._clone()
        qs.postFilter = self.__predicateBasedOn('_current_datetime_from')
        return qs

    def future(self):
        qs = self._clone()
        qs.postFilter = self.__predicateBasedOn('_future_datetime_from')
        return qs

    def past(self):
        qs = self._clone()
        qs.postFilter = self.__predicateBasedOn('_past_datetime_from')
        return qs

    def __predicateBasedOn(self, attribute):
        def predicate(item):
            # If used after byDay [ e.g. qry.byDay(from, to).upcoming() ] then
            # this will reject the whole days_events if just one event does not
            # match the predicate.
            for event in getattr(item, 'days_events', [item]):
                page = getattr(event, 'page', event)
                if not getattr(page, attribute, False):
                    return False
            return True
        return predicate

    def this(self):
        request = self.request
        class ThisIterable(ModelIterable):
            def __iter__(self):
                for page in super().__iter__():
                    yield ThisEvent(page.title, page, page.get_url(request))
        qs = self._clone()
        qs._iterable_class = ThisIterable
        return qs

    def authorized_q(self, request):
        PASSWORD = PageViewRestriction.PASSWORD
        LOGIN    = PageViewRestriction.LOGIN
        GROUPS   = PageViewRestriction.GROUPS
        KEY      = PageViewRestriction.passed_view_restrictions_session_key

        restrictions = PageViewRestriction.objects.all()
        passed = request.session.get(KEY, [])
        if passed:
            restrictions = restrictions.exclude(id__in=passed,
                                                restriction_type=PASSWORD)
        if request.user.is_authenticated:
            restrictions = restrictions.exclude(restriction_type=LOGIN)
        if request.user.is_superuser:
            restrictions = restrictions.exclude(restriction_type=GROUPS)
        else:
            membership = request.user.groups.all()
            if membership:
                restrictions = restrictions.exclude(groups__in=membership,
                                                    restriction_type=GROUPS)
        q = Q()
        for restriction in restrictions:
            q &= ~self.descendant_of_q(restriction.page, inclusive=True)
        return q

    def auth(self, request):
        self.request = request
        if request is None:
            return self
        else:
            return self.filter(self.authorized_q(request))

    # Possible Future feature redact unauthorized events??
    #def redact(self, request)

class EventPageForm(BorgPageForm):
    def clean(self):
        cleaned_data = super().clean()
        self._checkStartBeforeEnd(cleaned_data)
        return cleaned_data

    def _checkStartBeforeEnd(self, cleaned_data):
        startTime = getTimeFrom(cleaned_data.get('time_from'))
        endTime   = getTimeTo(cleaned_data.get('time_to'))
        if startTime > endTime:
            self.add_error('time_to', _("Event cannot end before it starts"))

# Cannot serialize: functools._lru_cache_wrapper object
# There are some values Django cannot serialize into migration files.
def _get_default_timezone():
    return timezone.get_default_timezone()

class EventBase(models.Model):
    class Meta:
        abstract = True

    uid = models.CharField(max_length=255, db_index=True, editable=False,
                           default=uuid4)
    category = models.ForeignKey(EventCategory,
                                 related_name="+",
                                 verbose_name=_("category"),
                                 on_delete=models.SET_NULL,
                                 blank=True, null=True)
    image = models.ForeignKey(get_image_model_string(),
                              null=True, blank=True,
                              related_name='+',
                              verbose_name=_("image"),
                              on_delete=models.SET_NULL)

    time_from = models.TimeField(_("start time"), null=True, blank=True)
    time_to = models.TimeField(_("end time"), null=True, blank=True)
    tz = TimeZoneField(verbose_name=_("time zone"),
                       default=_get_default_timezone)

    group_page  = models.ForeignKey(get_group_model_string(),
                                    null=True, blank=True,
                                    verbose_name=_("group page"),
                                    on_delete=models.SET_NULL)
    details  = RichTextField(_("details"), blank=True)
    location = models.CharField(_("location"), max_length=255, blank=True)
    website = models.URLField(_("website"), blank=True)

    # Init these variables to prevent template DEBUG messages
    # Yes, this is very ugly.  An alternative solution would be welcome.
    cancellation_details = None
    extra_information    = None
    postponed_from_when  = None

    search_fields = Page.search_fields + [
        index.SearchField('location'),
        index.SearchField('details'),
    ]

    content_panels1 = [
        FieldPanel('details', classname="full"),
        MapFieldPanel('location'),
        FieldPanel('website'),
    ]
    if getattr(settings, "JOYOUS_GROUP_SELECTABLE", False):
        content_panels1.append(PageChooserPanel('group_page'))

    # Anything inheriting from models.Model needs its own __init__ or
    # modeltranslation patch_constructor may break it
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def group(self):
        """
        The group this event belongs to.  Adding the event as a child of a
        group automatically assigns the event to that group.
        """
        retval = None
        parent = self.get_parent()
        Group = get_group_model()
        if issubclass(parent.specific_class, Group):
            retval = parent.specific
        if retval is None:
            retval = self.group_page
        return retval

    @property
    def _current_datetime_from(self):
        """
        The datetime this event will start or did start in the local timezone, or
        None if it is finished.
        """
        fromDt = self._getFromDt()
        toDt = self._getToDt()
        return fromDt if toDt >= timezone.localtime() else None

    @property
    def _future_datetime_from(self):
        """
        The datetime this event next starts in the local timezone, or None if
        in the past.
        """
        fromDt = self._getFromDt()
        return fromDt if fromDt >= timezone.localtime() else None

    @property
    def _past_datetime_from(self):
        """
        The datetime this event previously started in the local timezone, or
        None if it never did.
        """
        fromDt = self._getFromDt()
        return fromDt if fromDt < timezone.localtime() else None

    @property
    def _first_datetime_from(self):
        """
        The datetime this event first started in the local time zone, or None if
        it never did.
        """
        return self._getFromDt()

    @property
    def status(self):
        """
        The current status of the event (started, finished or pending).
        """
        now = timezone.localtime()
        if self._getToDt() < now:
           return "finished"
        elif self._getFromDt() < now:
           return "started"

    @property
    def status_text(self):
        """
        A text description of the current status of the event.
        """
        status = self.status
        if status == "finished":
            return _("This event has finished.")
        elif status == "started":
            return _("This event has started.")
        else:
            return ""

    @property
    def at(self):
        """
        A string describing what time the event starts (in the local time zone).
        """
        return timeFormat(self._getFromTime())

    @classmethod
    def _removeContentPanels(cls, *args):
        """
        Remove the panels and so hide the fields named.
        """
        remove = []
        for arg in args:
            if type(arg) is str:
                remove.append(arg)
            else:
                remove.extend(arg)
        cls.content_panels = _filterContentPanels(cls.content_panels, remove)

    def isAuthorized(self, request):
        """
        Is the user authorized for the requested action with this event?
        """
        restrictions = self.get_view_restrictions()
        if restrictions and request is None:
            return False
        else:
            return all(restriction.accept_request(request)
                       for restriction in restrictions)

    def get_context(self, request, *args, **kwargs):
        retval = super().get_context(request, *args, **kwargs)
        retval['themeCSS'] = getattr(settings, "JOYOUS_THEME_CSS", "")
        return retval

    def _getLocalWhen(self, date_from, date_to=None):
        """
        Returns a string describing when the event occurs (in the local time zone).
        """
        dateFrom, timeFrom = getLocalDateAndTime(date_from, self.time_from,
                                                 self.tz, dt.time.min)

        if date_to is not None:
            dateTo, timeTo = getLocalDateAndTime(date_to, self.time_to, self.tz)
        else:
            if self.time_to is not None:
                dateTo, timeTo = getLocalDateAndTime(date_from, self.time_to, self.tz)
            else:
                dateTo = dateFrom
                timeTo = None

        if dateFrom == dateTo:
            retval = _("{date} {atTime}").format(date=dateFormat(dateFrom),
                        atTime=timeFormat(timeFrom, timeTo, gettext("at ")))
        else:
            retval = _("{date} {atTime}").format(date=dateFormat(dateFrom),
                        atTime=timeFormat(timeFrom, prefix=gettext("at ")))
            retval = _("{dateTimeFrom} to {dateTo} {atTimeTo}").format(
                        dateTimeFrom=retval.strip(),
                        dateTo=dateFormat(dateTo),
                        atTimeTo=timeFormat(timeTo, prefix=gettext("at ")))
        return retval.strip()

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone) for the given date.
        """
        raise NotImplementedError()

    def _getFromDt(self):
        """
        Datetime that the event starts (in the local time zone).
        """
        # This is used by the default implementations of
        # _current_datetime_from, _future_datetime_from, _past_datetime_from,
        # and _first_datetime_from
        raise NotImplementedError()

    def _getToDt(self):
        """
        Datetime that the event ends (in the local time zone).
        """
        # This is used by the default implementation of
        # _current_datetime_from
        raise NotImplementedError()

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous Events API
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from contextlib import suppress
from functools import partial
from itertools import chain, groupby
from operator import attrgetter
from django.conf import settings
from django.core.exceptions import (MultipleObjectsReturned, ObjectDoesNotExist,
        PermissionDenied)
from django.utils.translation import gettext_lazy as _
from ..utils.weeks import week_of_month
from .event_base import EventsOnDay
from .one_off_events import SimpleEventPage, MultidayEventPage
from .recurring_events import (RecurringEventPage, MultidayRecurringEventPage,
        PostponementPage, RescheduleMultidayEventPage, ExtraInfoPage,
        CancellationPage, ClosedForHolidaysPage, ExtCancellationPage)

# ------------------------------------------------------------------------------
# API get functions
# ------------------------------------------------------------------------------
def getAllEventsByDay(request, fromDate, toDate, *, home=None, holidays=None):
    """
    Return all the events (under home if given) for the dates given, grouped by
    day.

    :param request: Django request object
    :param fromDate: starting date (inclusive)
    :param toDate: finish date (inclusive)
    :param home: only include events that are under this page (if given)
    :param holidays: the holidays that are celebrated for these dates
    :rtype: list of :class:`EventsOnDay <ls.joyous.models.events.EventsOnDay>` objects
    """
    qrys = [SimpleEventPage.events(request).byDay(fromDate, toDate),
            MultidayEventPage.events(request).byDay(fromDate, toDate),
            RecurringEventPage.events(request, holidays).byDay(fromDate, toDate),
            PostponementPage.events(request).byDay(fromDate, toDate)]
    # Cancellations and ExtraInfo pages are returned by RecurringEventPage.byDay
    if home is not None:
        qrys = [qry.descendant_of(home) for qry in qrys]
    evods = _getEventsByDay(fromDate, qrys, holidays)
    return evods

def getAllEventsByWeek(request, year, month, *, home=None, holidays=None):
    """
    Return all the events (under home if given) for the given month, grouped by
    week.

    :param request: Django request object
    :param year: the year
    :type year: int
    :param month: the month
    :type month: int
    :param home: only include events that are under this page (if given)
    :param holidays: the holidays that are celebrated for these dates
    :returns: a list of sublists (one for each week) each of 7 elements which are either None for days outside of the month, or the events on the day.
    :rtype: list of lists of None or :class:`EventsOnDay <ls.joyous.models.events.EventsOnDay>` objects
    """
    return _getEventsByWeek(year, month,
                            partial(getAllEventsByDay, request,
                                    home=home, holidays=holidays))

def getAllUpcomingEvents(request, *, home=None, holidays=None):
    """
    Return all the upcoming events (under home if given).

    :param request: Django request object
    :param home: only include events that are under this page (if given)
    :param holidays: holidays that may affect these events
    :rtype: list of ThisEvents
    """
    qrys = [SimpleEventPage.events(request).upcoming().this(),
            MultidayEventPage.events(request).upcoming().this(),
            RecurringEventPage.events(request, holidays).upcoming().this(),
            PostponementPage.events(request).upcoming().this(),
            ExtraInfoPage.events(request).exclude(extra_title="")
                            .upcoming().this(),
            CancellationPage.events(request).exclude(cancellation_title="")
                            .upcoming().this(),
            ExtCancellationPage.events(request).exclude(cancellation_title="")
                            .upcoming().this(),
            ClosedForHolidaysPage.events(request, holidays)
                            .exclude(cancellation_title="").upcoming().this()]
    if home is not None:
        qrys = [qry.descendant_of(home) for qry in qrys]
    events = sorted(chain.from_iterable(qrys), key=_getUpcomingSort())
    return events

def getGroupUpcomingEvents(request, group, holidays=None):
    """
    Return all the upcoming events that are assigned to the specified group.

    :param request: Django request object
    :param group: for this group page
    :param holidays: holidays that may affect these events
    :rtype: list of ThisEvents
    """
    if not hasattr(group, 'recurringeventpage_set'):
        # This is not a group page
        return []

    # Get events that are a child of a group page, or a postponement or extra
    # info a child of the recurring event child of the group
    rrEvents = RecurringEventPage.events(request, holidays)                  \
                                        .exclude(group_page=group)           \
                                        .upcoming().child_of(group).this()
    qrys = [SimpleEventPage.events(request).exclude(group_page=group)
                                        .upcoming().child_of(group).this(),
            MultidayEventPage.events(request).exclude(group_page=group)
                                        .upcoming().child_of(group).this(),
            rrEvents]
    for rrEvent in rrEvents:
        qrys += [PostponementPage.events(request).child_of(rrEvent.page)
                                         .upcoming().this(),
                 ExtraInfoPage.events(request).exclude(extra_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 CancellationPage.events(request).exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 ExtCancellationPage.events(request).exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 ClosedForHolidaysPage.events(request, holidays)
                                 .exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this()]

    # Get events that are linked to a group page, or a postponement or extra
    # info child of the recurring event linked to a group
    rrEvents = group.recurringeventpage_set(manager='events').auth(request)  \
                                 .hols(holidays).upcoming().this()
    qrys += [group.simpleeventpage_set(manager='events').auth(request)
                                 .upcoming().this(),
             group.multidayeventpage_set(manager='events').auth(request)
                                 .upcoming().this(),
             rrEvents]
    for rrEvent in rrEvents:
        qrys += [PostponementPage.events(request).child_of(rrEvent.page)
                                 .upcoming().this(),
                 ExtraInfoPage.events(request).exclude(extra_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 CancellationPage.events(request).exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 ExtCancellationPage.events(request).exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this(),
                 ClosedForHolidaysPage.events(request, holidays)
                                 .exclude(cancellation_title="")
                                 .child_of(rrEvent.page).upcoming().this()]
    events = sorted(chain.from_iterable(qrys), key=_getUpcomingSort())
    return events

def getAllPastEvents(request, *, home=None, holidays=None):
    """
    Return all the past events (under home if given).

    :param request: Django request object
    :param home: only include events that are under this page (if given)
    :param holidays: holidays that may affect these events
    :rtype: list of the ThisEvents
    """
    qrys = [SimpleEventPage.events(request).past().this(),
            MultidayEventPage.events(request).past().this(),
            RecurringEventPage.events(request, holidays).past().this(),
            PostponementPage.events(request).past().this(),
            ExtraInfoPage.events(request).exclude(extra_title="").past().this(),
            CancellationPage.events(request).exclude(cancellation_title="")
                            .past().this(),
            ExtCancellationPage.events(request).exclude(cancellation_title="")
                            .past().this(),
            ClosedForHolidaysPage.events(request, holidays)
                            .exclude(cancellation_title="")
                            .past().this()]
    if home is not None:
        qrys = [qry.descendant_of(home) for qry in qrys]
    events = sorted(chain.from_iterable(qrys),
                    key=attrgetter('page._past_datetime_from'), reverse=True)
    return events

def getEventFromUid(request, uid):
    """
    Get the event by its UID
    (raises PermissionDenied if we have no authority, ObjectDoesNotExist if it
    is not found).

    :param request: Django request object
    :param uid: iCal unique identifier
    :rtype: event page
    """
    events = []
    with suppress(ObjectDoesNotExist):
        events.append(SimpleEventPage.objects.get(uid=uid))
    with suppress(ObjectDoesNotExist):
        events.append(MultidayEventPage.objects.get(uid=uid))
    with suppress(ObjectDoesNotExist):
        events.append(RecurringEventPage.objects.get(uid=uid))
    # Exceptions do not have uids and are not returned by this function

    if len(events) == 1:
        if events[0].isAuthorized(request):
            return events[0]
        else:
            raise PermissionDenied("No authority for uid={}".format(uid))
    elif len(events) == 0:
        raise ObjectDoesNotExist("No event with uid={}".format(uid))
    else:
        raise MultipleObjectsReturned("Multiple events with uid={}".format(uid))

def getAllEvents(request, *, home=None, holidays=None):
    """
    Return all the events (under home if given).

    :param request: Django request object
    :param home: only include events that are under this page (if given)
    :param holidays: holidays that may affect these events
    :rtype: list of event pages
    """
    qrys = [SimpleEventPage.events(request).all(),
            MultidayEventPage.events(request).all(),
            RecurringEventPage.events(request, holidays).all()]
    # Does not return exceptions
    if home is not None:
        qrys = [qry.descendant_of(home) for qry in qrys]
    events = sorted(chain.from_iterable(qrys),
                    key=attrgetter('_first_datetime_from'))
    return events

# ------------------------------------------------------------------------------
# API UI functions
# ------------------------------------------------------------------------------
def removeContentPanels(*args):
    """
    Remove the panels and so hide the fields named.
    """
    remove = []
    for arg in args:
        if type(arg) is str:
            remove.append(arg)
        else:
            remove.extend(arg)

    SimpleEventPage._removeContentPanels(remove)
    MultidayEventPage._removeContentPanels(remove)
    RecurringEventPage._removeContentPanels(remove)
    MultidayRecurringEventPage._removeContentPanels(remove)
    PostponementPage._removeContentPanels(remove)
    RescheduleMultidayEventPage._removeContentPanels(remove)

# ------------------------------------------------------------------------------
# Private
# ------------------------------------------------------------------------------
def _getEventsByDay(date_from, eventsByDaySrcs, holidays):
    if holidays is None:
        holidays = {}
    evods = []
    day = date_from
    # TODO sortByTime could go up here
    # TODO would izip be better?
    for srcs in zip(*eventsByDaySrcs):
        days_events       = []
        continuing_events = []
        for src in srcs:
            days_events += src.days_events
            continuing_events += src.continuing_events
        def sortByTime(thisEvent):
            fromTime = thisEvent.page._getFromTime(atDate=day)
            if fromTime is None:
                fromTime = dt.time.max
            return fromTime
        days_events.sort(key=sortByTime)
        holiday = holidays.get(day)
        evods.append(EventsOnDay(day, holiday, days_events, continuing_events))
        day += dt.timedelta(days=1)
    return evods

def _getEventsByWeek(year, month, eventsByDaySrc):
    weeks = []
    firstDay = dt.date(year, month, 1)
    lastDay  = dt.date(year, month, calendar.monthrange(year, month)[1])
    def calcWeekOfMonth(evod):
        return week_of_month(evod.date)
    events = eventsByDaySrc(firstDay, lastDay)
    for weekOfMonth, group in groupby(events, calcWeekOfMonth):
        week = list(group)
        if len(week) < 7:
            padding = [None] * (7 - len(week))
            if weekOfMonth == 0:
                week = padding + week
            else:
                week += padding
        weeks.append(week)
    return weeks

def _getUpcomingSort():
    if getattr(settings, "JOYOUS_UPCOMING_INCLUDES_STARTED", False):
        return attrgetter('page._current_datetime_from')
    else:
        return attrgetter('page._future_datetime_from')

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test ClosedForHolidays Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from itertools import islice
from django.test import RequestFactory, TestCase
from django.contrib.auth.models import User
from wagtail.core.models import Page
from wagtail.tests.utils.form_data import rich_text
from holidays.holiday_base import HolidayBase
from ls.joyous.holidays import Holidays
from ls.joyous.models import CalendarPage
from ls.joyous.models import (RecurringEventPage, CancellationPage,
                              ExtCancellationPage)
from ls.joyous.models import ClosedForHolidaysPage, ClosedFor
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import DAILY, WEEKLY, MONTHLY, MO, WE, FR
from .testutils import freeze_timetz, datetimetz

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3(r3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug      = "test-meeting",
                                        title     = "Test Meeting",
                                        repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                               freq=WEEKLY,
                                                               byweekday=[MO,WE,FR]),
                                        time_from = dt.time(13),
                                        time_to   = dt.time(15,30),
                                        holidays = self.calendar.holidays)
        self.calendar.add_child(instance=self.event)
        self.closedHols = ClosedForHolidaysPage(owner = self.user,
                                                overrides = self.event,
                                                all_holidays = True,
                                                holidays = self.calendar.holidays)
        self.event.add_child(instance=self.closedHols)
        self.closedHols.save_revision().publish()

    def testCanCreateOnlyOne(self):
        self.assertFalse(ClosedForHolidaysPage.can_create_at(self.event))

    def testInit(self):
        self.assertEqual(self.closedHols.all_holidays, True)
        self.assertEqual(self.closedHols.title, "Closed for holidays")
        self.assertEqual(self.closedHols.local_title, "Closed for holidays")
        self.assertEqual(self.closedHols.slug,  "closed-for-holidays")

    def testGetEventsByDay(self):
        events = RecurringEventPage.events.hols(self.calendar.holidays)      \
                                   .byDay(dt.date(1989,1,1), dt.date(1989,1,31))
        self.assertEqual(len(events), 31)
        evod = events[0]
        self.assertEqual(evod.date, dt.date(1989,1,1))
        self.assertEqual(evod.holiday, "New Year's Day")
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 0)

    @freeze_timetz("1990-11-11 16:29:00")
    def testEventFutureExceptions(self):
        request = RequestFactory().get("/test")
        request.user = self.user
        request.session = {}
        exceptions = self.event._futureExceptions(request)
        self.assertEqual(len(exceptions), 1)
        c4h = exceptions[0]
        self.assertEqual(c4h.title, "Closed for holidays")
        self.assertEqual(c4h._future_datetime_from, datetimetz(1990,11,16,13,0))

    def testEventOccursOn(self):
        self.assertIs(self.event._occursOn(dt.date(1989, 1, 1)), False)
        self.assertIs(self.event._occursOn(dt.date(1989, 1, 2)), False)
        self.assertIs(self.event._occursOn(dt.date(1989, 1, 6)), True)
        self.assertIs(self.event._occursOn(dt.date(1989, 1, 16)), False)

    @freeze_timetz("2022-12-24")
    def testEventNextDate(self):
        self.assertEqual(self.event.next_date, dt.date(2022,12,28))

    @freeze_timetz("2023-01-31")
    def testEventPastDatetime(self):
        self.assertEqual(self.event._past_datetime_from, datetimetz(2023,1,27,13,0))

    @freeze_timetz("1990-10-11 16:29:00")
    def testGetUpcomingEvents(self):
        event = RecurringEventPage(slug      = "RST",
                                   title     = "Ruritania secret taxidermy",
                                   repeat    = Recurrence(dtstart=dt.date(1980,1,1),
                                                          freq=MONTHLY,
                                                          byweekday=[MO(1)]),
                                   time_from = dt.time(20))
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False,
                                           cancellation_title = "Closed for the holiday")
        closedHols.closed_for = [ ClosedFor(name="Wellington Anniversary Day"),
                                  ClosedFor(name="Auckland Anniversary Day"),
                                  ClosedFor(name="Nelson Anniversary Day"),
                                  ClosedFor(name="Taranaki Anniversary Day"),
                                  ClosedFor(name="Otago Anniversary Day"),
                                  ClosedFor(name="Southland Anniversary Day"),
                                  ClosedFor(name="South Canterbury Anniversary Day"),
                                  ClosedFor(name="Hawke's Bay Anniversary Day"),
                                  ClosedFor(name="Marlborough Anniversary Day"),
                                  ClosedFor(name="Canterbury Anniversary Day"),
                                  ClosedFor(name="Chatham Islands Anniversary Day"),
                                  ClosedFor(name="Westland Anniversary Day") ]
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        events = ClosedForHolidaysPage.events.hols(self.calendar.holidays)   \
                                      .exclude(cancellation_title="")        \
                                      .upcoming().this().descendant_of(event)
        self.assertEqual(len(events), 1)
        title, page, url  = events[0]
        self.assertEqual(title, "Closed for the holiday")
        self.assertEqual(page._future_datetime_from, datetimetz(1990,12,3,20,0))
        self.assertEqual(url, "/events/RST/closed-for-holidays/")

    @freeze_timetz("1990-10-11 16:29:00")
    def testGetPastEvents(self):
        events = ClosedForHolidaysPage.events.hols(self.calendar.holidays)   \
                                      .past().this()
        self.assertEqual(len(events), 1)
        title, page, url  = events[0]
        self.assertEqual(title, "")
        self.assertEqual(page._past_datetime_from, datetimetz(1990,9,24,13,0))
        self.assertEqual(url, "/events/test-meeting/closed-for-holidays/")

    @freeze_timetz("1990-10-11 16:29:00")
    def testWombatGetEventsByDay(self):
        event = RecurringEventPage(slug   = "UVW",
                                   title  = "Underwater viking wombats",
                                   repeat = Recurrence(dtstart=dt.date(1989,1,1),
                                                       freq=MONTHLY,
                                                       byweekday=[MO(1)]),
                                   time_from = dt.time(19))
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False,
                                           cancellation_title = "UVW Cancelled")
        closedHols.closed_for = [ ClosedFor(name="New Year's Day"),
                                  ClosedFor(name="Day after New Year's Day"),
                                  ClosedFor(name="Good Friday"),
                                  ClosedFor(name="Easter Monday"),
                                  ClosedFor(name="Christmas Day"),
                                  ClosedFor(name="Boxing Day") ]
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        events = RecurringEventPage.events.hols(self.calendar.holidays)      \
                                   .byDay(dt.date(1989,1,1), dt.date(1989,1,31))
        self.assertEqual(len(events), 31)
        evod = events[1]
        self.assertEqual(evod.date, dt.date(1989,1,2))
        self.assertEqual(evod.holiday, "Day after New Year's Day")
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        title, page, url = evod.all_events[0]
        self.assertEqual(title, "UVW Cancelled")
        self.assertEqual(page.title, "Closed for holidays")
        self.assertEqual(page.at, "7pm")
        self.assertEqual(url, "/events/UVW/closed-for-holidays/")

    def testClosedForDates(self):
        dates = list(islice(self.closedHols._closed_for_dates, 10))
        self.assertEqual(dates, [ dt.date(1989,1,2),
                                  dt.date(1989,1,16),
                                  dt.date(1989,1,23),
                                  dt.date(1989,1,30),
                                  dt.date(1989,2,6),
                                  dt.date(1989,3,13),
                                  dt.date(1989,3,20),
                                  dt.date(1989,3,24),
                                  dt.date(1989,3,27),
                                  dt.date(1989,6,5) ])

    def testGetMyDates(self):
        event = RecurringEventPage(slug      = "UVW",
                                   title     = "Underwater viking wombats",
                                   repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                          freq=MONTHLY,
                                                          byweekday=[MO(1)]),
                                   time_from = dt.time(19))
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False)
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        dates = list(islice(closedHols._getMyDates(), 10))
        self.assertEqual(dates, [])
        closedHols.holidays = self.calendar.holidays
        dates = list(islice(closedHols._getMyDates(), 10))
        self.assertEqual(dates, [])
        closedHols.closed_for = [ ClosedFor(name="Good Friday") ]
        # Good Friday is never going to fall on a Monday
        dates = list(islice(closedHols._getMyDates(), 10))
        self.assertEqual(dates, [])

    def testClosedOn(self):
        event = RecurringEventPage(slug      = "XYZ",
                                   title     = "Xylophone yacht zombies",
                                   repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                          freq=WEEKLY,
                                                          byweekday=[FR]),
                                   time_from = dt.time(19),
                                   holidays = self.calendar.holidays)
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False,
                                           cancellation_title = "XYZ Cancelled")
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        self.assertEqual(closedHols._closedOn(dt.date(1989, 3, 24)), False)
        closedHols.holidays = self.calendar.holidays
        self.assertEqual(closedHols._closedOn(dt.date(1989, 3, 24)), False)

    def testStatus(self):
        self.assertEqual(self.closedHols.status, "cancelled")
        self.assertEqual(self.closedHols.status_text, "Closed for holidays.")

    def testWhen(self):
        self.assertEqual(self.closedHols.when, "Closed on all holidays")

    def testWhenEver(self):
        event = RecurringEventPage(slug      = "XYZ",
                                   title     = "Xylophone yacht zombies",
                                   repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                          freq=WEEKLY,
                                                          byweekday=[FR]),
                                   time_from = dt.time(19),
                                   holidays = self.calendar.holidays)
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False,
                                           cancellation_title = "XYZ Cancelled",
                                           holidays = self.calendar.holidays)
        closedHols.closed_for = [ ClosedFor(name="Good Friday"),
                                  ClosedFor(name="Easter Monday") ]
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        self.assertEqual(closedHols.when, "Closed on Good Friday and Easter Monday")
        self.assertIs(event._occursOn(dt.date(1989, 3, 24)), False)

    # def testActualHolidaysClosed(self):
    #     event = RecurringEventPage(slug      = "Z1",
    #                                title     = "ZetaOne",
    #                                repeat    = Recurrence(dtstart=dt.date(2020,1,1),
    #                                                       freq=DAILY),
    #                                time_from = dt.time(19),
    #                                holidays = self.calendar.holidays)
    #     self.calendar.add_child(instance=event)
    #     closedHols = ClosedForHolidaysPage(owner = self.user,
    #                                        overrides = event,
    #                                        all_holidays = False,
    #                                        cancellation_title = "Z1 Cancelled",
    #                                        holidays = self.calendar.holidays)
    #     closedHols.closed_for = [ ClosedFor(name="New Year's Day"),
    #                               ClosedFor(name="New Year's Day (Observed)"),
    #                               ClosedFor(name="Labour Day"),
    #                               ClosedFor(name="Christmas Day"),
    #                               ClosedFor(name="Christmas Day (Observed)") ]
    #     event.add_child(instance=closedHols)
    #     closedHols.save_revision().publish()
    #     self.assertCountEqual(closedHols._actual_closed_for_names,
    #                           ["New Year's Day",
    #                            "New Year's Day (Observed)",
    #                            "Labour Day",
    #                            "Christmas Day",
    #                            "Christmas Day (Observed)"])

    def testAt(self):
        self.assertEqual(self.closedHols.at.strip(), "1pm")

    @freeze_timetz("1989-02-15")
    def testCurrentDt(self):
        # Taranaki Anniversary Day
        self.assertEqual(self.closedHols._current_datetime_from,
                         datetimetz(1989,3,13,13,0))

    @freeze_timetz("1989-02-15")
    def testFutureDt(self):
        # Taranaki Anniversary Day
        self.assertEqual(self.closedHols._future_datetime_from,
                         datetimetz(1989,3,13,13,0))

    @freeze_timetz("1989-02-15")
    def testPastDt(self):
        # Waitangi Day
        self.assertEqual(self.closedHols._past_datetime_from,
                         datetimetz(1989,2,6,13,0))

    @freeze_timetz("1989-02-15")
    def testFutureDtNoHolidays(self):
        self.closedHols.holidays = None
        self.assertEqual(self.closedHols._future_datetime_from,
                         ClosedForHolidaysPage.MAX_DATETIME)

    @freeze_timetz("1989-02-15")
    def testPastDtNoHolidays(self):
        self.closedHols.holidays = None
        self.assertEqual(self.closedHols._past_datetime_from,
                         ClosedForHolidaysPage.MIN_DATETIME)

    @freeze_timetz("1989-02-15")
    def testFutureDtCancelled(self):
        # a cancellation page trumps the holiday
        cancellation = CancellationPage(owner = self.user,
                                        overrides = self.event,
                                        except_date = dt.date(1989, 3, 13))
        self.event.add_child(instance=cancellation)
        self.assertEqual(self.closedHols._future_datetime_from,
                         datetimetz(1989,3,20,13,0))

    @freeze_timetz("1989-02-15")
    def testPastDtCancelled(self):
        # a cancellation page trumps the holiday
        cancellation = CancellationPage(owner = self.user,
                                        overrides = self.event,
                                        except_date = dt.date(1989, 2, 6))
        self.event.add_child(instance=cancellation)
        self.assertEqual(self.closedHols._past_datetime_from,
                         datetimetz(1989,1,30,13,0))

    @freeze_timetz("1989-02-15")
    def testPastDtShut(self):
        # a shutdown page trumps the holiday
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = self.event,
                                       cancelled_from_date = dt.date(1989, 1, 29),
                                       cancelled_to_date   = dt.date(1989, 2, 10))
        self.event.add_child(instance=shutdown)
        self.assertEqual(self.closedHols._past_datetime_from,
                         datetimetz(1989,1,23,13,0))

    @freeze_timetz("1989-02-15")
    def testFutureDtShut(self):
        # a shutdown page trumps the holiday
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = self.event,
                                       cancelled_from_date = dt.date(1989, 3, 12),
                                       cancelled_to_date   = dt.date(1989, 3, 20))
        self.event.add_child(instance=shutdown)
        self.assertEqual(self.closedHols._future_datetime_from,
                         datetimetz(1989,3,24,13,0))

    @freeze_timetz("1989-02-15")
    def testFutureDtShutUntilFurtherNotice(self):
        # a shutdown page trumps the holiday
        shutdown = ExtCancellationPage(owner = self.user,
                                       overrides = self.event,
                                       cancelled_from_date = dt.date(1989, 3, 12))
        self.event.add_child(instance=shutdown)
        self.assertIsNone(self.closedHols._future_datetime_from)

    @freeze_timetz("1989-02-15")
    def testFarFutureDt(self):
        event = RecurringEventPage(slug      = "UVW",
                                   title     = "Underwater viking wombats",
                                   repeat    = Recurrence(dtstart=dt.date(1989,1,1),
                                                          freq=MONTHLY,
                                                          bymonthday=23),
                                   time_from = dt.time(19))
        self.calendar.add_child(instance=event)
        closedHols = ClosedForHolidaysPage(owner = self.user,
                                           overrides = event,
                                           all_holidays = False)
        event.add_child(instance=closedHols)
        closedHols.save_revision().publish()
        dates = list(islice(closedHols._getMyDates(), 10))
        self.assertEqual(dates, [])
        closedHols.holidays = self.calendar.holidays
        dates = list(islice(closedHols._getMyDates(), 10))
        self.assertEqual(dates, [])
        closedHols.closed_for = [ ClosedFor(name="Good Friday") ]
        # Good Friday falls on 2038-04-23 but that is too far to calculate
        self.assertEqual(closedHols._future_datetime_from,
                         datetimetz(9999,12,29,19,0))

    def testGroup(self):
        self.assertIsNone(self.closedHols.group)

    def testOverridesRepeat(self):
        self.assertEqual(self.closedHols.overrides_repeat, self.event.repeat)

    def testGetContext(self):
        request = RequestFactory().get("/test")
        context = self.closedHols.get_context(request)
        self.assertIn('overrides', context)

    def testClosedForStr(self):
        xmas = ClosedFor(name="☧mas")
        self.assertEqual(str(xmas), "☧mas")

# ------------------------------------------------------------------------------
class WeatherDays(HolidayBase):
    def _populate(self, year):
        self[dt.date(year, 2, 2)]   = "Groundhog Day"
        self[dt.date(year, 7, 15)]  = "St Swithin's Day"
        self[dt.date(year, 7, 27)]  = "Sleepy Head Day"

class TestPageForm(TestCase):
    Form = ClosedForHolidaysPage.get_edit_handler().get_form_class()

    def setUp(self):
        holidays = Holidays(holidaySetting=None)
        holidays.register(WeatherDays())
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@bar.test', 's3(r3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.calendar.holidays = holidays
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug      = "committee-meeting",
                                        title     = "Committee Meeting",
                                        repeat    = Recurrence(dtstart=dt.date(2017,1,1),
                                                               freq=MONTHLY,
                                                               byweekday=[MO(1), MO(3)]),
                                        time_from = dt.time(13),
                                        time_to   = dt.time(15,30),
                                        holidays  = holidays)
        self.calendar.add_child(instance=self.event)

    def testValid(self):
        page = ClosedForHolidaysPage(owner = self.user,
                                     holidays = self.event.holidays)
        form = self.Form({'overrides':    self.event,
                          'all_holidays': False,
                          'closed_for':   ["St Swithin's Day",
                                           "Sleepy Head Day"],
                          'cancellation_title':   "Holiday",
                          'cancellation_details': rich_text("No meeting today!"), },
                         instance = page, parent_page = self.event)
        self.assertCountEqual(form.fields['closed_for'].choices,
                              [("Groundhog Day",    "Groundhog Day"),
                               ("St Swithin's Day", "St Swithin's Day"),
                               ("Sleepy Head Day",  "Sleepy Head Day")])
        self.assertEqual(form.initial['closed_for'], [])
        self.assertTrue(form.is_valid())         # is_valid() calls full_clean()
        self.assertDictEqual(form.errors, {})
        saved = form.save(commit=False)
        self.assertCountEqual(saved.closed, ["St Swithin's Day",
                                             "Sleepy Head Day"])

    def testDefaultHolidayChoices(self):
        page = ClosedForHolidaysPage(owner = self.user)
        form = self.Form({'overrides':    self.event},
                         instance = page, parent_page = self.event)
        self.assertCountEqual(form.fields['closed_for'].choices,
                              [(name, name) for name in (
                                  "New Year's Day",
                                  "Day after New Year's Day",
                                  "New Year's Day (Observed)",
                                  "Day after New Year's Day (Observed)",
                                  'Wellington Anniversary Day',
                                  'Auckland Anniversary Day',
                                  'Nelson Anniversary Day',
                                  'Waitangi Day',
                                  'Waitangi Day (Observed)',
                                  'Taranaki Anniversary Day',
                                  'Otago Anniversary Day',
                                  'Good Friday',
                                  'Easter Monday',
                                  'Southland Anniversary Day',
                                  'Anzac Day',
                                  'Anzac Day (Observed)',
                                  "Queen's Birthday",
                                  'South Canterbury Anniversary Day',
                                  "Hawke's Bay Anniversary Day",
                                  'Labour Day',
                                  'Marlborough Anniversary Day',
                                  'Canterbury Anniversary Day',
                                  'Chatham Islands Anniversary Day',
                                  'Westland Anniversary Day',
                                  'Christmas Day',
                                  'Boxing Day',
                                  'Christmas Day (Observed)',
                                  'Boxing Day (Observed)')])

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Forms
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.test import TestCase, override_settings
from django.contrib.auth.models import User
from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.core.models import Page
from ls.joyous.forms import FormDefender, FormClassOverwriteWarning
from ls.joyous.models import CalendarPage
from ls.joyous.models.one_off_events import MultidayEventPage as MEP
from ls.joyous.models.one_off_events import MultidayEventPageForm as MEPForm


# ------------------------------------------------------------------------------
class NewPageForm(WagtailAdminPageForm):
    def clean(self):
        # take a copy of cleaned_data so add_error does not remove fields
        cleaned_data = dict(super().clean())
        startDate = cleaned_data.get('date_from', dt.date.min)
        endDate   = cleaned_data.get('date_to', dt.date.max)
        msg = 'Choose a date in this Millennium'
        if startDate.year <= 2000:
            self.add_error('date_from', msg)
        if endDate.year <= 2000:
            self.add_error('date_to', msg)
        return cleaned_data

    def save(self, commit=True):
        buck = User.objects.create_user('buck', 'buck@nasa.gov', 'w!Lm@')
        page = super().save(commit=False)
        page.owner = buck
        if commit:
            page.save()
        return page

class NewMEPForm(MEPForm):
    pass

class Test(TestCase):
    def setUp(self):
        MEP._base_form_class = MEPForm
        MEP.get_edit_handler.cache_clear()
        if hasattr(MEPForm, 'assimilated_class'):
            del MEPForm.assimilated_class

    def tearDown(self):
        pass

    def testType(self):
        self.assertIs(type(MEP), FormDefender)

    def testNew(self):
        self.assertTrue('_base_form_class' in MEP.__dict__)
        self.assertFalse('base_form_class' in MEP.__dict__)

    def testGet(self):
        self.assertTrue(hasattr(MEP, 'base_form_class'))
        self.assertEqual(MEP.base_form_class, MEPForm)

    def testSetNone(self):
        with self.assertWarns(FormClassOverwriteWarning):
            MEP.base_form_class = None
        self.assertIs(MEP.base_form_class, None)

    def testSetNewForm(self):
        with self.assertWarns(FormClassOverwriteWarning):
            MEP.base_form_class = NewPageForm
        self.assertEqual(MEP.base_form_class, NewPageForm)

    def testSetNewFromNone(self):
        MEP._base_form_class = None
        MEP.base_form_class = NewPageForm
        self.assertEqual(MEP.base_form_class, NewPageForm)

    @override_settings(JOYOUS_DEFEND_FORMS=True)
    def testSetSubclass(self):
        MEP.base_form_class = NewMEPForm
        self.assertEqual(MEP.base_form_class, MEPForm)

    @override_settings(JOYOUS_DEFEND_FORMS=True)
    def testDefend(self):
        MEP.base_form_class = None
        self.assertIs(MEP.base_form_class, MEPForm)

    def testAssimilate(self):
        MEPForm.assimilate(NewPageForm)
        self.assertEqual(MEPForm.assimilated_class, NewPageForm)

    @override_settings(JOYOUS_DEFEND_FORMS=True)
    def testAssimilated(self):
        MEP.base_form_class = NewPageForm
        self.assertIs(MEP.base_form_class, MEPForm)
        Form = MEP.get_edit_handler().get_form_class()
        self.assertTrue(issubclass(Form, MEPForm))
        form = Form({'slug':      "C2C",
                     'title':     "Coast to Coast",
                     'date_from': "2021-02-12",
                     'date_to':   "2021-02-13",
                     'tz':        "Pacific/Auckland"})
        self.assertIsInstance(form.assimilated, NewPageForm)
        self.assertEqual(form.assimilated.fields.keys(), form.fields.keys())
        self.assertTrue(form.is_valid())
        self.assertDictEqual(form.errors, {})

    @override_settings(JOYOUS_DEFEND_FORMS=True)
    def testAssimilatedClean(self):
        MEP.base_form_class = NewPageForm
        Form = MEP.get_edit_handler().get_form_class()
        form = Form({'slug':      "C2C",
                     'title':     "Coast to Coast",
                     'date_from': "1987-02-08",
                     'date_to':   "1987-02-07",
                     'tz':        "Pacific/Auckland"})
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors,
                             {'date_from': ["Choose a date in this Millennium"],
                              'date_to':   ["Choose a date in this Millennium",
                                            "Event cannot end before it starts",]})

    @override_settings(JOYOUS_DEFEND_FORMS=True)
    def testAssimilatedSave(self):
        user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        calendar = CalendarPage(owner = user,
                                slug  = "events",
                                title = "Events")
        home = Page.objects.get(slug='home')
        home.add_child(instance=calendar)
        calendar.save_revision().publish()
        event = MEP(owner = user,
                    slug  = "allnighter",
                    title = "All Night",
                    date_from = dt.date(2012,12,31),
                    date_to   = dt.date(2013,1,1),
                    time_from = dt.time(23),
                    time_to   = dt.time(3))
        calendar.add_child(instance=event)
        event.save_revision().publish()
        MEP.base_form_class = NewPageForm
        Form = MEP.get_edit_handler().get_form_class()
        form = Form(instance=event, parent_page=calendar)
        page = form.save()
        self.assertEqual(page.owner.username, "buck")

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# Generated by Django 3.0.6 on 2020-07-14 10:39

from django.db import migrations, models
import django.db.models.deletion
import wagtail.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0016_closedfor_closedforholidayspage'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExtCancellationPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('cancellation_title', models.CharField(blank=True, help_text='Show in place of cancelled event (Leave empty to show nothing)', max_length=255, verbose_name='title')),
                ('cancellation_details', wagtail.core.fields.RichTextField(blank=True, help_text='Why was the event cancelled?', verbose_name='details')),
                ('cancelled_from_date', models.DateField(help_text='Cancelled from this date', verbose_name='From Date')),
                ('cancelled_to_date', models.DateField(blank=True, help_text='Cancelled to this date (Leave empty for "until further notice")', null=True, verbose_name='To Date')),
                ('overrides', models.ForeignKey(help_text='The recurring event that we are updating.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.RecurringEventPage', verbose_name='overrides')),
            ],
            options={
                'verbose_name': 'extended cancellation',
                'verbose_name_plural': 'extended cancellations',
                'default_manager_name': 'objects',
            },
            bases=('wagtailcore.page', models.Model),
        ),
    ]



# ------------------------------------------------------------------------------
# Joyous events models
# ------------------------------------------------------------------------------
import datetime as dt
from django.db import models
from django.db.models.query import ModelIterable
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from wagtail.core.models import Page
from wagtail.admin.edit_handlers import FieldPanel
from wagtail.images.edit_handlers import ImageChooserPanel

from ..utils.telltime import (todayUtc, getAwareDatetime, getLocalDatetime,
        getLocalDate, getLocalTime)
from ..utils.telltime import timeFormat
from ..edit_handlers import TimePanel
from ..forms import FormDefender
from .groups import get_group_model_string
from .event_base import (ThisEvent, EventsByDayList,
        EventManager, EventQuerySet, EventPageForm, EventBase)

# ------------------------------------------------------------------------------
# Helper types and constants
# ------------------------------------------------------------------------------
_1day  = dt.timedelta(days=1)
_2days = dt.timedelta(days=2)

# ------------------------------------------------------------------------------
# Event models
# ------------------------------------------------------------------------------
class SimpleEventQuerySet(EventQuerySet):
    def current(self):
        qs = super().current()
        return qs.filter(date__gte = todayUtc() - _1day)

    def future(self):
        qs = super().future()
        return qs.filter(date__gte = todayUtc() - _1day)

    def past(self):
        qs = super().past()
        return qs.filter(date__lte = todayUtc() + _1day)

    def byDay(self, fromDate, toDate):
        request = self.request
        class ByDayIterable(ModelIterable):
            def __iter__(self):
                evods = EventsByDayList(fromDate, toDate)
                for page in super().__iter__():
                    pageFromDate = getLocalDate(page.date,
                                                page.time_from, page.tz)
                    pageToDate   = getLocalDate(page.date,
                                                page.time_to, page.tz)
                    thisEvent = ThisEvent(page, url=page.get_url(request))
                    evods.add(thisEvent, pageFromDate, pageToDate)
                yield from evods

        qs = self._clone()
        qs._iterable_class = ByDayIterable
        return qs.filter(date__range=(fromDate - _2days, toDate + _2days))

class SimpleEventPage(EventBase, Page, metaclass=FormDefender):
    events = EventManager.from_queryset(SimpleEventQuerySet)()

    class Meta:
        verbose_name = _("event page")
        verbose_name_plural = _("event pages")
        default_manager_name = "objects"

    parent_page_types = ["joyous.CalendarPage",
                         "joyous.SpecificCalendarPage",
                         "joyous.GeneralCalendarPage",
                         get_group_model_string()]
    subpage_types = []
    base_form_class = EventPageForm

    date    = models.DateField(_("date"), default=dt.date.today)

    content_panels = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('date'),
        TimePanel('time_from'),
        TimePanel('time_to'),
        FieldPanel('tz'),
        ] + EventBase.content_panels1

    # Anything inheriting from models.Model needs its own __init__ or
    # modeltranslation patch_constructor may break it
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        return self._getLocalWhen(self.date)

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone).
        """
        return getLocalTime(self.date, self.time_from, self.tz)

    def _getFromDt(self):
        """
        Datetime that the event starts (in the local time zone).
        """
        return getLocalDatetime(self.date, self.time_from, self.tz)

    def _getToDt(self):
        """
        Datetime that the event ends (in the local time zone).
        """
        return getLocalDatetime(self.date, self.time_to, self.tz)

# ------------------------------------------------------------------------------
class MultidayEventQuerySet(EventQuerySet):
    def current(self):
        qs = super().current()
        return qs.filter(date_to__gte = todayUtc() - _1day)

    def future(self):
        qs = super().future()
        return qs.filter(date_from__gte = todayUtc() - _1day)

    def past(self):
        qs = super().past()
        return qs.filter(date_from__lte = todayUtc() + _1day)

    def byDay(self, fromDate, toDate):
        request = self.request
        class ByDayIterable(ModelIterable):
            def __iter__(self):
                evods = EventsByDayList(fromDate, toDate)
                for page in super().__iter__():
                    pageFromDate = getLocalDate(page.date_from,
                                                page.time_from, page.tz)
                    pageToDate   = getLocalDate(page.date_to,
                                                page.time_to, page.tz)
                    thisEvent = ThisEvent(page, url=page.get_url(request))
                    evods.add(thisEvent, pageFromDate, pageToDate)
                yield from evods

        qs = self._clone()
        qs._iterable_class = ByDayIterable
        return qs.filter(date_to__gte   = fromDate - _2days)   \
                 .filter(date_from__lte = toDate + _2days)

class MultidayEventPageForm(EventPageForm):
    def _checkStartBeforeEnd(self, cleaned_data):
        startDate = cleaned_data.get('date_from', dt.date.min)
        endDate   = cleaned_data.get('date_to', dt.date.max)
        if startDate > endDate:
            self.add_error('date_to', _("Event cannot end before it starts"))
        elif startDate == endDate:
            super()._checkStartBeforeEnd(cleaned_data)

class MultidayEventPage(EventBase, Page, metaclass=FormDefender):
    events = EventManager.from_queryset(MultidayEventQuerySet)()

    class Meta:
        verbose_name = _("multiday event page")
        verbose_name_plural = _("multiday event pages")
        default_manager_name = "objects"

    parent_page_types = ["joyous.CalendarPage",
                         "joyous.SpecificCalendarPage",
                         "joyous.GeneralCalendarPage",
                         get_group_model_string()]
    subpage_types = []
    base_form_class = MultidayEventPageForm

    date_from = models.DateField(_("start date"), default=dt.date.today)
    date_to = models.DateField(_("end date"), default=dt.date.today)

    content_panels = Page.content_panels + [
        FieldPanel('category'),
        ImageChooserPanel('image'),
        FieldPanel('date_from'),
        TimePanel('time_from'),
        FieldPanel('date_to'),
        TimePanel('time_to'),
        FieldPanel('tz'),
        ] + EventBase.content_panels1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def when(self):
        """
        A string describing when the event occurs (in the local time zone).
        """
        return self._getLocalWhen(self.date_from, self.date_to)

    def _getFromTime(self, atDate=None):
        """
        Time that the event starts (in the local time zone).
        """
        return getLocalTime(self.date_from, self.time_from, self.tz)

    def _getFromDt(self):
        """
        Datetime that the event starts (in the local time zone).
        """
        return getLocalDatetime(self.date_from, self.time_from, self.tz)

    def _getToDt(self):
        """
        Datetime that the event ends (in the local time zone).
        """
        return getLocalDatetime(self.date_to, self.time_to, self.tz)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
